import copy
import time
import pandas
import tqdm
from peegy.processing.pipe.definitions import InputOutputProcess
# from peegy.processing.pipe.definitions import Events
# from peegy.processing.pipe.io import GenericInputData
from peegy.processing.pipe.general import RemoveEvokedResponse
from peegy.processing.pipe.pipeline import PipePool
from peegy.definitions.tables import Tables
from peegy.processing.statistics.definitions import TestType
import numpy as np
import pandas as pd
import astropy.units as u
from peegy.processing.system.progress import ParallelTqdm
from joblib import delayed
import ray
import sys
from typing import List


class BootstrapTarget(object):
    """
    This class is used to define the target tables that will be taking for bootstrapping.
    """
    def __init__(self,
                 test_name: str | None = None,
                 group_by: List[str] | None = None,
                 target_values: List[str] | None = None
                 ):
        """

        :param table_name: What table should be targeted for bootstrapping
        :param group_by: column names (present in the target statistical table) that should be used to group
        the data that should be bootstrapped.
        :param target_values: column names with the statistics that should be bootstrapped (e.g. f)
        """
        self.test_name = test_name
        self.group_by = group_by
        self.target_values = target_values


class Bootstrap(InputOutputProcess):
    def __init__(self,
                 backend: str = 'joblib',
                 n_jobs: int | None = None,
                 origin_input_process: InputOutputProcess | None = None,
                 before_bootstrapping_input_process: InputOutputProcess | None = None,
                 at_each_bootstrap_do: PipePool | None = None,
                 n_events: int | None = None,
                 event_code: float = 1,
                 dur: u.Quantity = 100 * u.us,
                 min_time: type(u.Quantity) | None = None,
                 max_time: type(u.Quantity) | None = None,
                 epoch_length: type(u.Quantity) | None = None,
                 n_bootstraps: int = 1000,
                 bootstrap_targets: List[BootstrapTarget] | None = None,
                 alpha_level: float = 0.05,
                 one_sided: bool = True,
                 **kwargs) -> InputOutputProcess:
        """
        Bootstrap input pipeline. It assumed that the origin_input_process contains the continuous, un-epoched data.
        The events will be randomized and the pipeline passed in at_each_bootstrap_do will be run.
        The process of shuffling events and running the pipeline  will be repeated n_bootstraps times.
        The output table will contain the mean, median, standard deviation, lower confidence interval, upper confidence
        interval, sample size for the bootstrap and the p-value.
        Confidence intervals are based on the alpha_value passed.
        :param backend: which multiprocessing backend to use (ray or libjob)
        :param n_jobs: integer indicating the number of simultaneous jobs to be run.
        :param origin_input_process: The continuous, un-epoched data that will be bootstrap
        :param before_bootstrapping_input_process: Inputprocess to be run before bootstrap begins but after the
        original statistics have been computed. If None, by default the average evoked response will be removed.
        :param at_each_bootstrap_do: The pipeline that will be bootstrapped
        :param n_events: If not none, the number of random events will be n_events
        :param event_code: the event code that will be shuffled
        :param dur: duration of the shuffled event code
        :param min_time: minimum time allowed for shuffle events. If None, the timing of the first event will be used
        :param max_time: maximum time allowed for shuffle events. If None, the timing of the last event will be used
        :param epoch_length: reference duration of epochs. This is only used to ensure that the after the last event
        there will be time enough to ensure this epoch length
        :param n_bootstraps: how many times to bootstrap the pipeline
        :param bootstrap_targets: array of BootstrapTarget indicating what tables from the different pipeline elements
        should be keep for bootstrapping and what statistics.
        :param alpha_level: alpha level of confidence intervals
        :param one_sided: bool indicating whether alpha value is one_sided or not.
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(Bootstrap, self).__init__(input_process=None, **kwargs)
        self.backend = backend
        self.n_jobs = n_jobs
        self.origin_input_process = origin_input_process
        self.before_bootstrapping_input_process = before_bootstrapping_input_process
        self.at_each_bootstrap_do = at_each_bootstrap_do
        self.n_bootstraps = n_bootstraps
        self.n_events = n_events
        self.event_code = event_code
        self.dur = dur
        self.min_time = min_time
        self.max_time = max_time
        self.epoch_length = epoch_length
        self.alpha_level = alpha_level
        self.one_sided = one_sided
        self._original_events = None
        self.bootstrap_targets = bootstrap_targets
        self._origin_input_process = None
        self._at_each_bootstrap_do = None

    def transform_data(self):
        # the origin input_process is copied to ensure nothing is done to the original reference process
        self._at_each_bootstrap_do = copy.deepcopy(self.at_each_bootstrap_do)
        if self.origin_input_process is not None:
            self._origin_input_process = copy.deepcopy(self.origin_input_process)
        else:
            self._origin_input_process = copy.deepcopy(list(self._at_each_bootstrap_do.values())[0].input_process)
        # ensure that the reference process is used by the pipeline
        for _key, _process in list(self._at_each_bootstrap_do.items()):
            if _process.input_process.name == self._origin_input_process.name:
                self._at_each_bootstrap_do[_key].input_process = self._origin_input_process

        # first run pipeline to extract actual values to be tested against bootstrapped tables
        self._at_each_bootstrap_do.run(force=True)
        # extract reference tables
        ref_table = get_tables(pipeline=self._at_each_bootstrap_do, targets=self.bootstrap_targets)
        # get event information
        _n_events = self.n_events
        _min_time = self.min_time
        _max_time = self.max_time
        if _n_events is None:
            _n_events = self._origin_input_process.output_node.events.get_events_code(code=self.event_code).size
        if _min_time is None:
            _min_time = self._origin_input_process.output_node.events.get_events_time(code=self.event_code).min()
        if _max_time is None:
            _max_time = self._origin_input_process.output_node.events.get_events_time(code=self.event_code).max()
        # apply whatever we ask before we bootstrap. Normally we will remove the evoked data
        if self.before_bootstrapping_input_process is not None:
            self.before_bootstrapping_input_process.input_process = self._origin_input_process
        else:
            self.before_bootstrapping_input_process = RemoveEvokedResponse(
                input_process=self._origin_input_process,
                event_code=self.event_code,
                post_stimulus_interval=self.epoch_length)

        self.before_bootstrapping_input_process.run()
        self._origin_input_process.output_node.data = self.before_bootstrapping_input_process.output_node.data
        # generate average epoch length if not given to prevent small estimations from random events
        _epoch_length = self.epoch_length
        if self.epoch_length is None:
            _epoch_length = np.mean(
                np.diff(self._origin_input_process.output_node.events.get_events_time(code=self.event_code)))

        # bootstrap pipeline
        start_time = time.time()
        bootstrapped_tables = self.bootstrap_pipeline(backend=self.backend,
                                                      n_jobs=self.n_jobs,
                                                      event_code=self.event_code,
                                                      dur=self.dur,
                                                      epoch_length=_epoch_length,
                                                      min_time=_min_time,
                                                      max_time=_max_time,
                                                      n_events=_n_events,
                                                      )
        end_time = time.time()
        print('Bootstrapping took {:1.3} seconds'.format(end_time - start_time))
        # compute statistics
        print('Generating statistical table')
        boots_df = table_list_to_pd(tables=bootstrapped_tables, targets=self.bootstrap_targets)
        ref_table = ref_table.rename(columns={'value': 'statistic_value'})
        columns_left = set(boots_df.keys()).difference(['value'])
        columns_right = set(ref_table.keys()).difference(['statistic_value'])
        common_cols = list(columns_left.intersection(columns_right))
        bootstrapped_and_ref_tables = pd.merge(boots_df,
                                               ref_table,
                                               how='left',
                                               on=common_cols,
                                               )

        df_boots = bootstrapped_and_ref_tables.groupby(list(set(bootstrapped_and_ref_tables.keys()).difference(
            ['value', 'statistic_value'])), as_index=False, dropna=False).apply(
            lambda x: pd.Series({'stats': get_bootstrap_estimates(samples=x['value'],
                                                                  statistic=x['statistic_value'],
                                                                  alpha_level=self.alpha_level,
                                                                  one_sided=self.one_sided)}))

        (df_boots['statistic_value'],
         df_boots['mean'],
         df_boots['median'],
         df_boots['std'],
         df_boots['ci_lower'],
         df_boots['ci_upper'],
         df_boots['sample_size'],
         df_boots['p_value']) = zip(*df_boots.stats)
        df_boots['one_sided'] = self.one_sided
        df_boots['data_source_bootstrap'] = df_boots['data_source']
        df_boots = df_boots.drop(['stats', 'data_source'], axis=1)
        df_boots['n_epochs'] = _n_events
        # join boot table and reference table
        self.output_node.statistical_tests = Tables(table_name=TestType.bootstrap,
                                                    data=df_boots,
                                                    data_source=self.name)

    def bootstrap_pipeline(self,
                           n_jobs: int | None = None,
                           epoch_length: type(u.Quantity) | None = None,
                           n_events: int | None = None,
                           event_code: float = 1,
                           dur: u.Quantity = 100 * u.us,
                           min_time: type(u.Quantity) | None = None,
                           max_time: type(u.Quantity) | None = None,
                           backend: str = 'joblib'
                           ):
        if n_jobs is None:
            n_jobs = 1
        bootstrapped_tables = []
        if backend == 'ray':
            # multiprocessing bootstrap via ray
            ray.shutdown()
            if not ray.is_initialized():
                if 'sphinx' in sys.modules:
                    # this is a protection for when building with sphinx
                    sys.stdout.fileno = lambda: False
                ray.init(log_to_driver=False)
            ref_data = ray.put(self._origin_input_process.output_node.data.copy())
            self._origin_input_process.input_node.data = None
            self._origin_input_process.output_node.data = None
            for _, _value in self._at_each_bootstrap_do.items():
                _value.input_node.data = None
                _value.output_node.data = None
            reference_pool_id = self._at_each_bootstrap_do
            target_tables = self.bootstrap_targets
            ref_process = self._origin_input_process

            subs = np.maximum(self.n_bootstraps // n_jobs, 1)
            n_boots = np.array([subs] * n_jobs)
            _current_number = np.sum(n_boots)
            if _current_number < self.n_bootstraps:
                n_boots[-1] = n_boots[-1] + self.n_bootstraps - _current_number
            result_ids = [ray_worker_bootstrap_pool.options(num_cpus=n_jobs).remote(
                ref_process=ref_process,
                ref_data=ref_data,
                pipe_pool=reference_pool_id,
                bootstrap_targets=target_tables,
                event_code=event_code,
                dur=dur,
                epoch_length=epoch_length,
                min_time=min_time,
                max_time=max_time,
                n_boots=n_boots[_i],
                n_events=n_events) for _i in range(n_jobs)]
            results = [result for result in tqdm.tqdm(to_iterator(result_ids),
                                                      total=len(result_ids),
                                                      desc='Bootstrapping/worker',
                                                      colour='blue',
                                                      miniters=0)]
            ray.shutdown()
            for _r in results:
                bootstrapped_tables = bootstrapped_tables + _r
        elif backend == 'joblib':
            # multiprocessing bootstrap via joblib
            ref_data = self._origin_input_process.output_node.data.copy()
            if self._origin_input_process.input_node is not None:
                self._origin_input_process.input_node.data = None

            if self._origin_input_process.output_node is not None:
                self._origin_input_process.output_node.data = None
            for _, _value in self._at_each_bootstrap_do.items():
                _value.input_node.data = None
                _value.output_node.data = None
            reference_pool_id = self._at_each_bootstrap_do
            target_tables = self.bootstrap_targets
            # ref_process = copy.deepcopy(self._origin_input_process)
            ref_process = self._origin_input_process

            subs = np.maximum(self.n_bootstraps // n_jobs, 1)
            n_boots = np.array([subs] * n_jobs)
            _current_number = np.sum(n_boots)
            if _current_number < self.n_bootstraps:
                n_boots[-1] = n_boots[-1] + self.n_bootstraps - _current_number

            results = ParallelTqdm(total=n_jobs,
                                   desc='Bootstrapping/worker',
                                   n_jobs=n_jobs,
                                   mmap_mode='r',
                                   colour='blue',
                                   # backend='threading'
                                   )(
                delayed(worker_bootstrap_pool)(
                    ref_process=ref_process,
                    ref_data=ref_data,
                    pipe_pool=reference_pool_id,
                    bootstrap_targets=target_tables,
                    event_code=self.event_code,
                    epoch_length=epoch_length,
                    min_time=min_time,
                    max_time=max_time,
                    n_boots=n_boots[_i],
                    n_events=n_events
                ) for _i in range(n_jobs))
            for _r in results:
                bootstrapped_tables = bootstrapped_tables + _r
        return bootstrapped_tables


def to_iterator(obj_ids):
    while obj_ids:
        done, obj_ids = ray.wait(obj_ids)
        _result = ray.get(done[0])
        yield _result


@ray.remote
def ray_worker_bootstrap_pool(ref_process: InputOutputProcess | None = None,
                              ref_data: type(np.array) | None = None,
                              pipe_pool: PipePool | None = None,
                              bootstrap_targets: List[BootstrapTarget] | None = None,
                              epoch_length: type(u.Quantity) | None = None,
                              n_events: int | None = None,
                              event_code: float = 1,
                              dur: u.Quantity = 100 * u.us,
                              min_time: type(u.Quantity) | None = None,
                              max_time: type(u.Quantity) | None = None,
                              n_boots: int | None = None
                              ):
    tables = worker_bootstrap_pool(ref_process=ref_process,
                                   ref_data=ref_data,
                                   pipe_pool=pipe_pool,
                                   bootstrap_targets=bootstrap_targets,
                                   epoch_length=epoch_length,
                                   n_events=n_events,
                                   event_code=event_code,
                                   dur=dur,
                                   min_time=min_time,
                                   max_time=max_time,
                                   n_boots=n_boots)
    return tables


def worker_bootstrap_pool(ref_process: InputOutputProcess | None = None,
                          ref_data: type(np.array) | None = None,
                          pipe_pool: PipePool | None = None,
                          bootstrap_targets: List[BootstrapTarget] | None = None,
                          epoch_length: type(u.Quantity) | None = None,
                          n_events: int | None = None,
                          event_code: float = 1,
                          dur: u.Quantity = 100 * u.us,
                          min_time: type(u.Quantity) | None = None,
                          max_time: type(u.Quantity) | None = None,
                          n_boots: int | None = None
                          ):
    ref_process.output_node.data = ref_data
    _local_pipe_pool = PipePool()
    for _key, _value in pipe_pool.items():
        _parameters = _value.get_input_parameters()
        if _parameters['input_process'].name == ref_process.name:
            _parameters['input_process'] = ref_process
        if _parameters['input_process'].name in _local_pipe_pool.keys():
            _parameters['input_process'] = _local_pipe_pool[_parameters['input_process'].name]
        _local_pipe_pool[_key] = type(_value)(**_parameters)

    bootstrapped_tables = []
    print(n_boots)
    for _i in tqdm.tqdm(range(n_boots),
                        total=n_boots,
                        desc='worker-bootstrapping',
                        colour='blue',
                        miniters=0):

        ref_process.output_node.randomize_events(event_code=event_code,
                                                 epoch_length=epoch_length,
                                                 min_time=min_time,
                                                 max_time=max_time,
                                                 dur=dur,
                                                 n_events=n_events
                                                 )

        _local_pipe_pool.run(force=True)
        _current_tables = catch_tables(pipeline=_local_pipe_pool, targets=bootstrap_targets)
        bootstrapped_tables = bootstrapped_tables + _current_tables
    return bootstrapped_tables


def get_bootstrap_estimates(samples: type(np.array) | None = None,
                            statistic: type(np.array) | None = None,
                            alpha_level: float = 0.05,
                            one_sided=True):
    # remove nan that may be present
    samples = np.array(samples[~pd.isna(samples)])
    statistic_value = np.unique(statistic)
    assert statistic_value.size == 1
    sample_size = samples.shape[0]
    sorted_samples = np.sort(samples, axis=0)
    mean = np.mean(samples, axis=0)
    median = np.median(samples, axis=0)
    std = np.std(samples, axis=0)
    _alpha = alpha_level
    if not one_sided:
        _alpha = alpha_level / 2.0

    _i_upper = np.minimum(int(samples.shape[0] * (1 - _alpha)), samples.shape[0])
    _i_lower = np.maximum(int(samples.shape[0] * _alpha), 0)
    ci_lower = sorted_samples[_i_lower]
    ci_upper = sorted_samples[_i_upper]
    diff = np.abs(statistic_value - sorted_samples)
    _idx_match = np.argmin(diff, axis=0) + 1
    p_value = 1 - _idx_match / sample_size
    return [statistic_value[0], mean, median, std, ci_lower, ci_upper, sample_size, p_value]


def get_tables(pipeline: PipePool | None = None,
               targets: List[BootstrapTarget] | None = None):
    output = pd.DataFrame()
    for _source, _value in pipeline.items():
        for _target in targets:
            for _key, df in _value.output_node.statistical_tests.items():
                if _target.test_name in df['test_name'].values:
                    _columns = list(set(df.keys()) &
                                    set(_target.target_values))
                    if len(_columns):
                        df = df[_target.group_by + _target.target_values + ['data_source', 'test_name']]
                        melted_df = pd.melt(df,
                                            ignore_index=False,
                                            id_vars=set(set(df.keys())).difference(_target.target_values),
                                            value_vars=_target.target_values,
                                            var_name='statistic',
                                            value_name='value')
                        output = pandas.concat([output, melted_df], ignore_index=True)
    return output


def catch_tables(pipeline: PipePool | None = None,
                 targets: List[BootstrapTarget] | None = None):
    output = []
    for _source, _value in pipeline.items():
        for _target in targets:
            for _key, df in _value.output_node.statistical_tests.items():
                if _target.test_name in df['test_name'].values:
                    _columns = list(set(df.keys()) &
                                    set(_target.target_values))
                    if len(_columns):
                        output.append(df)
    return output


def table_list_to_pd(tables: List[pd.DataFrame] | None = None,
                     targets: List[BootstrapTarget] | None = None):
    output = pd.DataFrame()
    for _target in targets:
        df = pd.concat([_t for _t in tables if _target.test_name in _t['test_name'].values])
        df = df[_target.group_by + _target.target_values + ['data_source', 'test_name']]
        melted_df = pd.melt(df, ignore_index=False,
                            id_vars=set(set(df.keys())).difference(_target.target_values),
                            value_vars=_target.target_values,
                            var_name='statistic')
        output = pandas.concat([output, melted_df], ignore_index=True)
    return output



import logging
import os
from struct import Struct
import astropy.units as u
import numpy as np
from peegy.definitions.channel_definitions import ChannelItem
from peegy.definitions.events import SingleEvent, Events
from peegy.tools.units.unit_tools import set_default_unit
from typing import List
log = logging.getLogger()
__author__ = 'jundurraga'
"These scripts are based on MNE-python and Neuroscan data structure http://paulbourke.net/dataformats/eeg/"


class CNTEventType1(object):
    def __init__(self,
                 stim_type,
                 key_board,
                 key_pad_accept,
                 offset):

        self.stim_type = stim_type
        self.key_board = key_board
        self.key_pad_accept = key_pad_accept
        self.offset = offset


class CNTEventType2(object):
    def __init__(self,
                 stim_type,
                 key_board,
                 key_pad_accept,
                 offset,
                 type,
                 code,
                 latency,
                 epoch_event,
                 accept2,
                 accuracy
                 ):
        self.stim_type = stim_type
        self.key_board = key_board
        self.key_pad_accept = key_pad_accept
        self.offset = offset
        self.type = type
        self.code = code
        self.latency = latency
        self.epoch_event = epoch_event
        self.accept2 = accept2
        self.accuracy = accuracy


class CNTEventType3(object):
    def __init__(self,
                 stim_type,
                 key_board,
                 key_pad_accept,
                 offset,
                 type,
                 code,
                 latency,
                 epoch_event,
                 accept2,
                 accuracy
                 ):
        self.stim_type = stim_type
        self.key_board = key_board
        self.key_pad_accept = key_pad_accept
        self.offset = offset
        self.type = type
        self.code = code
        self.latency = latency
        self.epoch_event = epoch_event
        self.accept2 = accept2
        self.accuracy = accuracy


class DeviceEventChannel:
    bdf_event_channel = 'Status'
    edf_event_channel = 'EDF Annotations'
    bdf_event_annotations = 'BDF Annotations'
    auto = 'auto'


def read_cnt_header(file_name: str):
    """Read the cnt header.
     http://paulbourke.net/dataformats/eeg/
    """
    data_offset = 900  # Size of the 'SETUP' header.
    with open(file_name, "rb", buffering=0) as f:
        f.seek(21)
        _subject_id = f.read(20).strip().decode('ascii', errors='ignore').replace('\x00', '')
        f.seek(121)
        subject_name = f.read(20).strip().decode('ascii', errors='ignore').replace('\x00', '').split()
        _last_name = subject_name[0] if len(subject_name) > 0 else ''
        _first_name = subject_name[-1] if len(subject_name) > 0 else ''
        f.seek(2, 1)
        _sex = f.read(1).strip().decode('ascii', errors='ignore').replace('\x00', '')
        _handedness = f.read(1).strip().decode('ascii', errors='ignore').replace('\x00', '')
        f.seek(205)
        _session_label = f.read(20).strip().decode('ascii', errors='ignore').replace('\x00', '')
        _session_date = f.read(10).strip().decode('ascii', errors='ignore').replace('\x00', '')
        _session_time = f.read(12).strip().decode('ascii', errors='ignore').replace('\x00', '')
        f.seek(353)
        n_events = np.fromfile(f, dtype="<u4", count=1).item()
        f.seek(370)
        n_channels = np.fromfile(f, dtype="<u2", count=1).item()
        f.seek(376)
        sampling_rate = np.fromfile(f, dtype="<u2", count=1).item() * u.Hz
        f.seek(438)
        lowpass_toggle = np.fromfile(f, "i1", count=1).item()
        highpass_toggle = np.fromfile(f, "i1", count=1).item()

        f.seek(864)
        n_samples = np.fromfile(f, dtype="<u4", count=1).item()
        _event_table_position = 886
        f.seek(_event_table_position)
        (event_offset,) = np.frombuffer(f.read(4), dtype="<i4")

        f.seek(869)
        low_cutoff = np.fromfile(f, dtype="f4", count=1).item()
        f.seek(2, 1)
        high_cutoff = np.fromfile(f, dtype="f4", count=1).item()
        n_bytes = 4

        f.seek(890)
        continuous_seconds = np.fromfile(f, dtype="<f4", count=1).item()
        events_size = os.path.getsize(file_name) - event_offset
        if events_size <= 0:  # no events
            data_size = n_samples * n_channels
        else:
            data_size = os.path.getsize(file_name) - (data_offset + 75 * n_channels) - events_size

        if n_samples == 0 or data_size // (n_samples * n_channels) not in [2, 4]:
            n_bytes = 2
            n_samples = data_size // (n_bytes * n_channels)
            # annotations = _read_annotations_cnt(input_fname, data_format="int16")
            # if len(annotations) and annotations.onset[-1] * sampling_rate > n_samples:
            #     n_bytes = 4
            #     n_samples = n_samples_header
        else:
            n_bytes = data_size // (n_samples * n_channels)

        # channel offset (blocks per channel)
        _channel_offset = np.fromfile(f, dtype="<i4", count=1).item()
        if _channel_offset > 1:
            _channel_offset = _channel_offset // n_bytes
        else:
            _channel_offset = 1

        if lowpass_toggle != 1:
            low_cutoff = None
        if highpass_toggle != 1:
            high_cutoff = None

        header = {'file_name': file_name,
                  'subject_id': _subject_id,
                  'first_name': _first_name,
                  'last_name': _last_name,
                  'sex': _sex,
                  'handedness': _handedness,
                  'session_label': _session_label,
                  'session_date': _session_date,
                  'n_channels': n_channels,
                  'lowpass_toggle': lowpass_toggle,
                  'highpass_toggle': highpass_toggle,
                  'low_cutoff': low_cutoff,
                  'high_cutoff': high_cutoff,
                  'file_size': os.path.getsize(file_name),
                  'start_date': _session_date,
                  'start_time': _session_time,
                  'bytes_in_header': data_offset,
                  'fs': sampling_rate,
                  'channel_offset': _channel_offset,
                  'n_bytes': n_bytes,
                  'n_samples': n_samples,
                  'has_events': events_size > 0,
                  'event_offset': event_offset,
                  'n_events': n_events,
                  'duration': n_samples / sampling_rate,
                  'continuous_seconds': continuous_seconds}
        (cals,
         baselines,
         sensitivities,
         channels,
         pos) = (list(), list(), list(), list(), list())

        ch = []
        for ch_idx in range(n_channels):
            f.seek(data_offset + 75 * ch_idx)
            ch_name = f.read(10).strip().decode('ascii', errors='ignore').replace('\x00', '')
            ch.append(ch_name)

            f.seek(data_offset + 75 * ch_idx + 19)
            xy = np.fromfile(f, dtype="f4", count=2)
            xy[1] *= -1  # invert y-axis
            pos.append(xy)
            f.seek(data_offset + 75 * ch_idx + 47)
            # baselines corrections
            baselines.append(np.fromfile(f, dtype="i2", count=1).item())

            f.seek(data_offset + 75 * ch_idx + 59)
            sensitivity = np.fromfile(f, dtype="f4", count=1).item()
            sensitivities.append(sensitivity)

            f.seek(data_offset + 75 * ch_idx + 71)
            cal = np.fromfile(f, dtype="f4", count=1).item()
            cals.append(cal * sensitivity / 204.8)  # calibration factor (micro Volts)
        for ch_name, (_pos_x, _pos_y) in zip(ch, pos):
            channels.append(ChannelItem(label=ch_name,
                                        idx=ch_idx,
                                        x=_pos_x,
                                        y=_pos_y,
                                        ))
        header['channels'] = np.array(channels)
        header['pos'] = np.array(pos)
        header['baselines'] = np.array(baselines)
        header['sensitivity'] = np.array(sensitivities)
        header['cals'] = np.array(cals)

    return header


def _get_event_parser(event_type):
    if event_type == 1:
        event_maker = CNTEventType1
        struct_pattern = "<HBcl"
    elif event_type == 2:
        event_maker = CNTEventType2
        struct_pattern = "<HBclhhfccc"
    elif event_type == 3:
        event_maker = CNTEventType3
        struct_pattern = "<HBclhhfccc"  # Same as event type 2
    else:
        raise ValueError(f"unknown CNT even type {event_type}")

    def parser(buffer):
        struct = Struct(struct_pattern)
        for chunk in struct.iter_unpack(buffer):
            yield event_maker(*chunk)

    return parser


def get_data(header,
             channels_idx: np.array = np.array([]),
             ini_time: u.quantity.Quantity = 0 * u.s,
             end_time: u.quantity.Quantity | None = None):
    fs_data = set_default_unit(header['fs'], u.Hz)
    ini_time = set_default_unit(ini_time, u.s)
    end_time = set_default_unit(end_time, u.s)

    n_channels = header['n_channels']
    channel_offset = header['channel_offset']
    baselines = header['baselines']
    cals = header['cals']
    n_bytes = header["n_bytes"]
    n_samples = header['n_samples']
    dtype = "<i4" if n_bytes == 4 else "<i2"
    chunk_size = channel_offset * n_channels
    if not channels_idx.size:
        channels_idx = np.arange(0, header['channels'].size).astype(int)
    # The data is divided into blocks of samples / channel.
    # channel_offset determines the amount of successive samples.
    # Offset is used to align the data because start can be in
    # the middle of these blocks.
    ini_sample = np.round(ini_time * fs_data).astype(int)
    if end_time is None:
        end_time = np.floor(header['duration'])
    else:
        end_time = np.minimum(end_time, header['duration'])
    end_sample = np.minimum(np.round(end_time * fs_data).astype(int), n_samples)
    data_left = (end_sample - ini_sample) * n_channels
    # Read up to 100 MB of data at a time, block_size is in data samples
    block_size = ((int(100e6) // n_bytes) // chunk_size) * chunk_size
    block_size = min(data_left, block_size)
    s_offset = ini_sample % channel_offset
    data = np.zeros((end_sample - ini_sample, n_channels))
    with open(header['file_name'], "rb", buffering=0) as f:
        f.seek(900 + n_channels * (75 + (ini_sample - s_offset) * n_bytes))
        for sample_start in np.arange(0, data_left, block_size) // n_channels:
            # Earlier comment says n_samples is unreliable, but I think it
            # is because it needed to be changed to unsigned int
            # See: PR #12393
            sample_stop = sample_start + min(
                (
                    n_samples,
                    block_size // n_channels,
                    data_left // n_channels - sample_start,
                )
            )
            n_samps = sample_stop - sample_start
            # In case channel offset and start time do not align perfectly,
            # extra sample sets are read here to cover the desired time
            # window. The whole (up to 100 MB) block is read at once and
            # then reshaped to (n_channels, n_samples).
            extra_samps = (
                chunk_size
                if (s_offset != 0 or n_samps % channel_offset != 0)
                else 0
            )
            if s_offset >= (channel_offset / 2):  # Extend at the end.
                extra_samps += chunk_size
            count = n_samps // channel_offset * chunk_size + extra_samps
            n_chunks = count // chunk_size
            samps = np.fromfile(f, dtype=dtype, count=count)
            samps = samps.reshape((n_chunks, n_channels, channel_offset), order="C")

            # Intermediate shaping to chunk sizes.
            block = np.zeros((channel_offset * n_chunks, n_channels))
            for set_idx, row in enumerate(samps):  # Final shape.
                block_slice = slice(
                    set_idx * channel_offset, (set_idx + 1) * channel_offset
                )
                block[block_slice, :n_channels] = row.T
            data[sample_start: sample_stop, channels_idx] = block[s_offset: n_samps + s_offset, channels_idx]
            print('reading {:}'.format((sample_stop/fs_data).to(u.s)))
    data = data[:, channels_idx]
    data -= baselines[channels_idx][None, :]
    data *= cals[channels_idx][None, :]
    data = data * u.uV
    # get events
    events = get_events(header=header,
                        ini_time=ini_time,
                        end_time=end_time)
    return data, events, channels_idx


def get_events(header,
               ini_time: u.quantity.Quantity = 0 * u.s,
               end_time: u.quantity.Quantity | None = None):
    # http://paulbourke.net/dataformats/eeg/
    events = None
    fs_data = set_default_unit(header['fs'], u.Hz)
    ini_time = set_default_unit(ini_time, u.s)
    end_time = set_default_unit(end_time, u.s)
    n_samples = header['n_samples']
    ini_sample = np.round(ini_time * fs_data).astype(int)
    if end_time is None:
        end_time = np.floor(header['duration'])
    else:
        end_time = np.minimum(end_time, header['duration'])
    end_sample = np.minimum(np.round(end_time * fs_data).astype(int), n_samples)
    if header['has_events']:
        teeg_parser = Struct("<Bll")
        with open(header['file_name'], "rb", buffering=0) as f:
            f.seek(header['event_offset'])
            (event_type, total_length, offset) = teeg_parser.unpack(f.read(teeg_parser.size))

        event_parser = _get_event_parser(event_type=event_type)

        with open(header['file_name'], "rb", buffering=0) as f:
            f.seek(header['event_offset'] + 9)  # the real table stats at +9
            buffer = f.read(total_length)

        orig_events = list(event_parser(buffer))
        events = _event_times(header=header,
                              original_events=orig_events,
                              ini_sample=ini_sample,
                              end_sample=end_sample)
    return events


def _event_times(header: dict | None = None,
                 original_events: List[object] | None = None,
                 ini_sample: int = 0,
                 end_sample: int | None = None) -> list:
    n_bytes = header['n_bytes']
    n_channels = header['n_channels']
    fs = header['fs']
    if end_sample is None:
        end_sample = np.inf
    events = []
    for _ev in original_events:
        offset = _ev.offset
        if isinstance(_ev, CNTEventType3):
            offset = offset * n_bytes * n_channels
        event_time = offset - 900 - (75 * n_channels)
        event_time = event_time // (n_channels * n_bytes)
        event_time = event_time - 1
        bad = False
        if "xd0" in str(_ev.key_pad_accept):
            bad = False
        if "xc0" in str(_ev.key_pad_accept):
            bad = True
        # Extract the 4-bit fields
        # Upper nibble (4 bits) currently not used
        # accept = (event.KeyPad_Accept[0] & 0xF0) >> 4
        # Lower nibble (4 bits) keypad button press
        keypad = _ev.key_pad_accept[0] & 0x0F
        description = None
        if str(keypad) != "0":
            description = f"KeyPad Response {keypad}"
        elif _ev.key_board != 0:
            description = f"Keyboard Response {_ev.key_board}"
        _event_time = (event_time - ini_sample) / fs
        _original_time = event_time / fs
        if (_event_time >= 0) and (_original_time <= end_sample / fs):
            events.append(SingleEvent(time_pos=_event_time,
                                      code=_ev.stim_type,
                                      bad_event=bad,
                                      description=description))
        if _original_time > end_sample / fs:
            break
    return Events(events)

    #     # Check to see if there are any button presses
    #     description = []
    #     for event in my_events:

    #
    #     description = np.array(description)
    #
    #     onset, duration, description = _update_bad_span_onset(
    #         accept_reject, onset / sfreq, duration, description
    #     )
    #     return Annotations(
    #         onset=onset, duration=duration, description=description, orig_time=None
    #     )



"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
from peegy.processing.tools.template_generator.auditory_waveforms import abr
from peegy.tools.signal_generator.noise_functions import generate_modulated_noise
import numpy as np
import matplotlib.pyplot as plt
import astropy.units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

# create synthetic data
fs = 20000.0 * u.Hz
duration = 1000 * u.ms
n_channels = 1
source, _ = abr(fs=fs, time_length=duration)

white_noise = generate_modulated_noise(fs=fs.value,
                                       duration=duration.to(u.s).value,
                                       f_noise_low=100,
                                       f_noise_high=10000,
                                       attenuation=0,
                                       n_channels=n_channels)

pink_noise = generate_modulated_noise(fs=fs.value,
                                      duration=duration.to(u.s).value,
                                      f_noise_low=100,
                                      f_noise_high=10000,
                                      attenuation=3,
                                      n_channels=n_channels)

brown_noise = generate_modulated_noise(fs=fs.value,
                                       duration=duration.to(u.s).value,
                                       f_noise_low=100,
                                       f_noise_high=10000,
                                       attenuation=6,
                                       n_channels=n_channels)

psd_pink_noise = np.abs(np.fft.rfft(pink_noise, axis=0)) ** 2.0
psd_brown_noise = np.abs(np.fft.rfft(brown_noise, axis=0)) ** 2.0
psd_white_noise = np.abs(np.fft.rfft(white_noise, axis=0)) ** 2.0
freqs = np.fft.rfftfreq(pink_noise.shape[0], 1 / fs)
fig = plt.figure()
ax = fig.add_subplot(111)
ax.plot(freqs, 10 * np.log10(psd_pink_noise / psd_pink_noise.max()), color='pink', label='Pink')
ax.plot(freqs, 10 * np.log10(psd_brown_noise / psd_brown_noise.max()), color='brown', label='Brown')
ax.plot(freqs, 10 * np.log10(psd_white_noise / psd_white_noise.max()), color='black', label='White')
ax.set_xscale('log')
ax.set_ylabel("PSD")
ax.set_xlabel("Frequency")
ax.grid(True)
ax.legend()
ax.set_xlim(100, 10000)
ax.set_ylim(-50, 2)
plt.show()



"""
.. _tut-assr-no-layout-test-sim:

########################################################
ASSR Without using a topographic layout (Simulated)
########################################################

In this example we simulate an ASSR recorded without any specific layout.
The significance of the response is assessed using the Hotelling's T2 test.

"""
# Enable below for interactive backend
# import matplotlib
# if 'qt5agg' in matplotlib.backends.backend_registry.list_builtin():
#     matplotlib.use('qt5agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Events, Domain
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels, RegressOutEOG, ReSampling
from peegy.processing.pipe.epochs import AverageEpochsFrequencyDomain, EpochData, AverageEpochs
from peegy.processing.pipe.plot import PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np
import peegy.definitions.default_paths as default_paths
# %%
# Generate some data
# ==================================
# We generate some auditory steady-state response (ASSR)


fs = 512.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 41.0 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
# here we pick some random frequencies to test statistical detection
random_frequencies = np.unique(np.random.rand(10)*30) * u.Hz
test_frequencies = np.concatenate(([assr_frequency], random_frequencies))
template_waveform, _ = aep(fs=fs)
n_channels = 8
event_times = np.arange(0, 360.0, 1/assr_frequency.to(u.Hz).value) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           snr=0.05,
                           include_eog_events=True,
                           event_times=event_times,
                           event_code=1.0,
                           figures_folder=default_paths.test_figures_path,
                           figures_subset_folder='assr_no_layout_test')
reader.run()
# %%
# Resize events
# ============================
# Now we keep events at intervals that correspond to our desired epoch length


events = reader.output_node.events.get_events(code=1)
# skip events to preserve only those at each epoch point
_new_events = Events(events=events[0:-1:int(epoch_length * assr_frequency)])
reader.output_node.events = _new_events

# %%
# Start the pipeline
# ============================
# Some processing to obtain clean epochs


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['CH_0'],
                                       invert_polarity=False)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['eog_removed'] = RegressOutEOG(pipeline['channel_cleaned'],
                                        ref_channel_labels=['EOG1'])
pipeline['down_sampled'] = ReSampling(pipeline['eog_removed'],
                                      new_sampling_rate=256. * u.Hz)
pipeline['time_filtered_data'] = FilterData(pipeline['down_sampled'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=60.0 * u.Hz)
pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False,
                                    post_stimulus_interval=epoch_length)
pipeline.run()

# %%
# Compute and plot components in the frequency-domain
# ---------------------------------------------------
# Spatial filter is applied in the frequency-domain

pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          projection_domain=Domain.frequency,
                                                          return_figures=True)
pipeline.run()

# %%
# Average epochs in time- and frequency-domain
# ----------------------------------------------
# We compute the average and simultaneously get statistical tests on the test_frequencies

pipeline['fft_ave'] = AverageEpochsFrequencyDomain(pipeline['dss_time_epochs'],
                                                   test_frequencies=test_frequencies)
pipeline['time_ave'] = AverageEpochs(pipeline['dss_time_epochs'])
pipeline.run()

# %%
# Generate figures
# ------------------
# Now we run plot the average waveforms


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['fft_ave'],
                                                   plot_x_lim=[0, 60],
                                                   plot_y_lim=[0, 6],
                                                   return_figures=True)
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['time_ave'],
                                                   times=np.array([0, 1 / assr_frequency.to(u.Hz).value]),
                                                   plot_x_lim=[0, epoch_length.to(u.s).value],
                                                   plot_y_lim=[-6, 6],
                                                   return_figures=True)
pipeline.run()

# %%
# Save results to a database
# ---------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ASSR'}

database_path = reader.input_node.paths.test_path.joinpath('assr_no_layout_test_data.sqlite')
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



from scipy.signal import hilbert
import astropy.units as u
import numpy as np
from scipy import signal
import matplotlib.pyplot as plt


def phase_value(v1, v2):
    vv1 = np.array([np.real(v1), np.imag(v1)]).T
    vv2 = np.array([np.real(v2), np.imag(v2)]).T
    cosine_angle = np.dot(vv1, vv2) / (np.linalg.norm(vv1) * np.linalg.norm(vv2))
    current_angle = np.arccos(cosine_angle)
    return current_angle


reference_frequency = 40 * u.Hz
half_width = 10 * u.ms
fs = 400 * u.Hz
M = int(10 * half_width * fs)
time = np.arange(M) / fs
s = (half_width.to(u.s) * fs)
w = reference_frequency * (2 * s * np.pi) / fs
wavelet = signal.morlet2(M, s, w)
plt.plot(time, abs(wavelet) / np.max(abs(wavelet)))
plt.plot(time, np.real(wavelet) / np.max(abs(wavelet)))
plt.show()

fs = 512 * u.Hz
reference_frequency = 80 * u.Hz
M = 100
s = 16.0
w = reference_frequency * (2 * s * np.pi) / fs
wavelet = signal.morlet2(M, s, w.value)
time = np.arange(0, 512) / fs
ini_phase = np.pi/2 * u.rad
reference_signal = (np.cos(2 * np.pi * reference_frequency * u.rad * time[:, None] - 0*ini_phase) +
                    1j * np.sin(2 * np.pi * reference_frequency * u.rad * time[:, None] - 0*ini_phase))

reference_signal = reference_signal * np.ones((reference_signal.shape[0], 10))
for phase in np.linspace(0, np.pi, 10) * u.rad:
    data = (np.cos(2 * np.pi * u.rad * reference_frequency * time - ini_phase)[:, None] +
            0.01 * np.random.randn(time.size, 1))

    data[data.shape[0] // 2::] = (np.cos(2 * np.pi * u.rad * reference_frequency * time[data.shape[0] // 2::] -
                                         ini_phase + phase)[:, None] +
                                  0.01 * np.random.randn(time[data.shape[0] // 2::].size, 1))

    y_filter = np.zeros(data.shape, dtype=complex)
    if data.ndim == 1:
        y_filter = signal.cwt(data.value, signal.morlet2, widths=[s], w=w)
    if data.ndim == 2:
        for _i in range(y_filter.shape[1]):
            y_filter[:, _i] = signal.cwt(data[:, _i].value, signal.morlet2,
                                         widths=[s], w=w)
    if data.ndim == 3:
        for _i in range(y_filter.shape[1]):
            for _j in range(y_filter.shape[2]):
                y_filter[:, _i, _j] = signal.cwt(data[:, _i, _j].value, signal.morlet2,
                                                 widths=[s], w=w)
    y_filter = hilbert(data, axis=0)
    # signal_instantaneous_phase = np.unwrap(np.angle(y_filter), axis=0) *
    # u.rad * np.ones((signal_instantaneous_phase.shape[0], 10))
    # plt.plot(time, signal_instantaneous_phase * 180 / np.pi, label='signal_phase')
    plt.figure(1)
    plt.plot(time, data, label='signal')
    plt.plot(time, np.real(reference_signal), label='signal')
    # plt.show()
    reference_instantaneous_phase = np.unwrap(np.angle(reference_signal), axis=0)
    plt.plot(time, reference_instantaneous_phase * 180 / np.pi, label='reference_phase')
    # if data.ndim <= 2:
    #     relative_instantaneous_phase = reference_instantaneous_phase - signal_instantaneous_phase
    # if data.ndim == 3:
    #     relative_instantaneous_phase = reference_instantaneous_phase[:, None] - signal_instantaneous_phase

    # relative_instantaneous_phase_2 = phase(y_filter, reference_signal)
    # M = 10
    # s = 1.0
    # w = 1 * reference_frequency * (2 * s * np.pi) / fs
    # wavelet = signal.morlet(M, s, w)
    # y_filter = signal.cwt(data.value.squeeze(), signal.morlet2, widths=[s], w=w).squeeze()
    # y_filter = signal.convolve(data.squeeze(), wavelet, 'same')[:, None]
    # y_filter = hilbert(np.real(y_filter), axis=0)
    # y_filter = signal.cwt(np.flip(y_filter, axis=0), signal.morlet2, widths=[s], w=w)[:, None]
    phase_vectorized = np.vectorize(phase_value)
    relative_instantaneous_phase_3 = np.unwrap(phase_vectorized(y_filter, reference_signal.value), axis=0)
    # mul = y_filter * reference_signal
    # dot = (np.real(mul) + np.imag(mul)) / (np.linalg.norm(np.real(mul)) * np.linalg.norm(np.imag(mul)))
    # relative_instantaneous_phase_2 = np.unwrap(np.arccos(dot), axis=0)
    plt.figure(2)
    # plt.plot(time, relative_instantaneous_phase * 180 / np.pi, label='hilbert')
    plt.plot(time, relative_instantaneous_phase_3.squeeze() * 180 / np.pi, label='projection_hilbert')
    # plt.plot(time, relative_instantaneous_phase_3 * 180 / np.pi, label='projection_morlet')
    plt.legend()
plt.show()
plt.show()



from __future__ import annotations
import pandas as pd


class Tables(dict):
    def __init__(self, table_name: str | None = None,
                 data: pd.DataFrame | None = None,
                 data_source: str | None = None):
        super(Tables, self).__init__()
        self._data_source = data_source
        if table_name is not None and data is not None:
            assert data_source is not None
            if 'data_source' not in data.keys():
                data['data_source'] = data_source
            self[table_name] = data

    def __setitem__(self, key, value):
        assert isinstance(value, pd.DataFrame)
        # key = self.ensure_unique_name(label=key)
        super(Tables, self).__setitem__(key, value)
        value.name = key

    def table_names(self):
        return self.keys()

    def append(self, input_table: Tables | None = None):
        for _key in input_table.keys():
            if _key in self.keys():
                self[_key] = pd.concat([self[_key], input_table[_key]], ignore_index=True)
            else:
                self[_key] = input_table[_key]

    def tables(self):
        return self

    def get_data(self, table_name: str | None = None):
        out = None
        if table_name in self.table_names():
            out = self[table_name]
        return out



import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
from peegy.processing.pipe.definitions import InputOutputProcess
from peegy.processing.tools.epochs_processing_tools import et_get_spatial_filtering, et_apply_spatial_filtering
from peegy.processing.tools.eeg_epoch_operators import et_unfold
from peegy.definitions.channel_definitions import Domain, ChannelItem
from peegy.processing.tools.epochs_processing_tools import et_ica_epochs
from peegy.processing.tools.eeg_epoch_operators import et_fold
from peegy.processing.pipe.epochs import AverageEpochs, AverageEpochsFrequencyDomain
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.definitions.tables import Tables
from peegy.processing.tools.filters.spatial_filtering.definitions import FilterType, DSSData
import numpy as np
import pandas as pd
import os
import copy
import astropy.units as u
from peegy.tools.units.unit_tools import set_default_unit
from PyQt5.QtCore import QLibraryInfo
os.environ["QT_QPA_PLATFORM_PLUGIN_PATH"] = QLibraryInfo.location(QLibraryInfo.PluginsPath)


class SpatialFilter(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 sf_join_frequencies: type(np.array) | None = None,
                 demean_data: bool = True,
                 weight_data: bool = True,
                 weighted_frequency_domain: bool = False,
                 weight_across_epochs: bool = True,
                 keep0: int | None = None,
                 keep1: float = 1e-9,
                 perc0: float = .99,
                 perc1: float | None = None,
                 block_size: int = 10,
                 n_tracked_points: int | None = None,
                 delta_frequency: u.Quantity = 5 * u.Hz,
                 n_jobs: int = 1,
                 **kwargs):
        """
        This class will create a spatial filter based on the data. If sf_join_frequencies are passed, the spatial
        filter will use a biased function based on the covariance of those frequencies only.
        :param input_process:  InputOutputProcess Class
        :param sf_join_frequencies: numpy array indicating the biased frequencies (useful for steady-state responses)
        :param demean_data: if true, data will be demeaned prior filter estimation
        :param weight_data: if true, dss filter will be estimated using weights
        :param weighted_frequency_domain: boll indicating if the weghts are compute in the time or frequency domain
        :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
        are computed within epoch (1 / variance across time)
        :param keep0: integer controlling  whitening of unbiased components in DSS. This integer value represent the
        number of components to keep.
        :param keep1: float controlling  whitening of unbiased components in DSS. This value will remove components
        below keep1 which is relative to the maximum eigen value.
        :param perc0: float (between 0 and 1) controlling whitening of unbiased components in DSS.
        This value will preserve components that explain the percentage of variance.
        :param perc1: float (between 0 and 1) controlling the number of biased components kept in DSS.
        This value will preserve the components of the biased PCA analysis that explain the percentage of variance.
        :param block_size: integer indicating the number of trials that would be used to estimate the weights
        :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
        :param delta_frequency: frequency size around each sf_join_frequency to estimate noise
        :param n_jobs: number of CPUs to compute FFT
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(SpatialFilter, self).__init__(input_process=input_process, **kwargs)
        self.sf_join_frequencies = set_default_unit(sf_join_frequencies, u.Hz)
        self.demean_data = demean_data
        self.weight_data = weight_data
        self.weighted_frequency_domain = weighted_frequency_domain
        self.weight_across_epochs = weight_across_epochs
        self.pwr0 = None
        self.pwr1 = None
        self.n_components_rotation_1 = None
        self.n_components_rotation_2 = None
        self.n_possible_components = None
        self.cov = None
        self.keep0 = keep0
        self.keep1 = keep1
        self.perc0 = perc0
        self.perc1 = perc1
        self.block_size = block_size
        self.n_tracked_points = n_tracked_points
        self.delta_frequency = delta_frequency
        self.weights = None
        self.n_jobs = n_jobs

    def transform_data(self):
        # compute spatial filter
        z, pwr0, pwr1, cov, n_0, n_1, weights = et_get_spatial_filtering(
            epochs=self.input_node.data,
            fs=self.input_node.fs,
            sf_join_frequencies=self.sf_join_frequencies,
            demean_data=self.demean_data,
            weight_data=self.weight_data,
            weighted_frequency_domain=self.weighted_frequency_domain,
            weight_across_epochs=self.weight_across_epochs,
            weights=self.input_node.w,
            keep0=self.keep0,
            keep1=self.keep1,
            perc0=self.perc0,
            perc1=self.perc1,
            block_size=self.block_size,
            n_tracked_points=self.n_tracked_points,
            delta_frequency=self.delta_frequency,
            n_jobs=self.n_jobs
        )
        self.output_node.data = z * u.dimensionless_unscaled
        self.pwr0 = pwr0
        self.pwr1 = pwr1
        self.cov = cov
        self.n_components_rotation_1 = n_0
        self.n_components_rotation_2 = n_1
        self.n_possible_components = self.input_node.data.shape[1]
        self.weights = weights


class ApplySpatialFilter(InputOutputProcess):
    def __init__(self,
                 input_process: SpatialFilter | None = None,
                 sf_components: np.array = np.array([]),
                 sf_thr: float = 0.8,
                 **kwargs):
        """
        This class will apply an SpatialFilter to the data. If sf_components are paased, only those will be used to
        filter the data
        :param input_process: an SpatialFilter InputOutputProcess Class
        :param sf_components: numpy array of integers with indexes of components to be kept
        :param sf_thr: float indicating the percentage of explained variance by components to be kept (this is used when
        sf_components is empty)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ApplySpatialFilter, self).__init__(input_process=input_process, **kwargs)
        self.sf_components = sf_components
        self.sf_thr = sf_thr
        self.kept_components = None

    def transform_data(self):
        filtered_data, components_idx = et_apply_spatial_filtering(z=self.input_node.data,
                                                                   pwr0=self.input_process.pwr0,
                                                                   pwr1=self.input_process.pwr1,
                                                                   cov_1=self.input_process.cov,
                                                                   sf_components=self.sf_components,
                                                                   sf_thr=self.sf_thr)
        self.output_node.data = filtered_data
        self.kept_components = components_idx


class CreateAndApplySpatialFilter(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 sf_join_frequencies: np.array([u.Quantity]) = None,
                 sf_components=np.array([]),
                 sf_thr: float = 0.8,
                 keep0: int | None = None,
                 keep1: float = 1e-9,
                 perc0: float = 1.0,
                 perc1: float | None = None,
                 test_frequencies: np.array([u.Quantity]) = None,
                 delta_frequency: u.Quantity = 5 * u.Hz,
                 block_size: int = 10,
                 n_tracked_points: int | None = None,
                 demean_data: bool = True,
                 weight_data: bool = True,
                 weighted_frequency_domain: bool = False,
                 weight_across_epochs: bool = True,
                 projection_domain: Domain = Domain.time,
                 components_to_plot: np.array = np.arange(0, 10),
                 plot_projections: bool = True,
                 plot_power: bool = True,
                 plot_x_lim: type([float, float]) = None,
                 plot_y_lim: type([float, float]) = None,
                 user_naming_rule: str | None = None,
                 fig_format: str = '.png',
                 return_figures: bool = False,
                 save_to_file: bool = True,
                 n_jobs: int = 1,
                 **kwargs):
        """
        This class will create and apply a spatial filter to the data.
        :param input_process: an InputOutputProcess Class
        :param sf_join_frequencies: numpy array indicating the biased frequencies (useful for steady-state responses)
        :param sf_components: numpy array of integers with indexes of components to be kept
        :param sf_thr: float indicating the percentage of explained variance by components to be kept (this is used when
        sf_components is empty)
        :param keep1: float controlling  whitening of unbiased components in DSS. This value will remove components
        below keep1 which is relative to the maximum eigen value.
        :param perc0: float (between 0 and 1) controlling whitening of unbiased components in DSS.
        This value will preserve components that explain the percentage of variance.
        :param perc1: float (between 0 and 1) controlling the number of biased components kept in DSS.
        This value will preserve the components of the biased PCA analysis that explain the percentage of variance.
        :param test_frequencies: frequency array to compute statistics in the frequency-domain in the component space.
        :param delta_frequency: frequency size around each sf_join_frequency to estimate noise
        :param demean_data: If true, filter will be created with demeaned data
        :param weight_data: if true, weighted average will be use to create the filter
        :param weighted_frequency_domain: boll indicating if the weghts are compute in the time or frequency domain
        :param weight_across_epochs: if True, weights are computed across epochs (as in Elbeling 1984) otherwise weights
        are computed within epoch (1 / variance across time)
        :param projection_domain: indicate the domain in which component will be projected in plots
        :param components_to_plot: numpy array indicating which components will be plotted.
        :param plot_projections: if true, components will be projected to the original space and plots will be generated
        :param plot_power: if true both bias and unbiased component's power as well as their rations will be plotted
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param fig_format: string indicating the format of the output figure (e.g. '.png' or '.pdf')
        :param return_figures: bool indicating if figure handle should be returned in self.figures
        :param save_to_file: bool indicating whether figure should be saved to file
        :param block_size: number of trials that will be stacked together to estimate the residual noise
        :param n_tracked_points: number of equally spaced points over time used to estimate residual noise and weights
        n_tracked_points
        :param n_jobs: number of CPUs for fft
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(CreateAndApplySpatialFilter, self).__init__(input_process=input_process, **kwargs)
        self.sf_join_frequencies = set_default_unit(sf_join_frequencies, u.Hz)
        self.sf_components = sf_components
        self.sf_thr = sf_thr
        self.keep0 = keep0
        self.keep1 = keep1
        self.perc0 = perc0
        self.perc1 = perc1
        self.p_ratio = None
        self.pwr0 = None
        self.pwr1 = None
        self.kept_components = None
        self.test_frequencies = test_frequencies
        self.delta_frequency = delta_frequency
        self.block_size = block_size
        self.n_tracked_points = n_tracked_points
        self.weights = None
        self.components_to_plot = components_to_plot
        self.plot_projections = plot_projections
        self.plot_power = plot_power
        self.projection_domain = projection_domain
        self.demean_data = demean_data
        self.weight_data = weight_data
        self.weighted_frequency_domain = weighted_frequency_domain
        self.weight_across_epochs = weight_across_epochs
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.user_naming_rule = user_naming_rule
        self.fig_format = fig_format
        self.return_figures = return_figures
        self.save_to_file = save_to_file
        self.n_jobs = n_jobs
        self.__kwargs = kwargs

    def transform_data(self):
        figures = None
        if self.test_frequencies is None:
            self.test_frequencies = self.sf_join_frequencies
        if self.return_figures:
            figures = []
        spatial_filter = SpatialFilter(input_process=self.input_process,
                                       sf_join_frequencies=self.sf_join_frequencies,
                                       demean_data=self.demean_data,
                                       weight_data=self.weight_data,
                                       weighted_frequency_domain=self.weighted_frequency_domain,
                                       weight_across_epochs=self.weight_across_epochs,
                                       keep0=self.keep0,
                                       keep1=self.keep1,
                                       perc0=self.perc0,
                                       perc1=self.perc1,
                                       block_size=self.block_size,
                                       n_tracked_points=self.n_tracked_points,
                                       delta_frequency=self.delta_frequency,
                                       n_jobs=self.n_jobs
                                       )
        spatial_filter.run()
        filtered_data = ApplySpatialFilter(input_process=spatial_filter,
                                           sf_components=self.sf_components,
                                           sf_thr=self.sf_thr,
                                           **self.__kwargs)
        filtered_data.run()

        self.output_node = filtered_data.output_node
        self.p_ratio = spatial_filter.pwr1 / spatial_filter.pwr0
        self.pwr0 = spatial_filter.pwr0
        self.pwr1 = spatial_filter.pwr1
        self.kept_components = filtered_data.kept_components
        self.weights = spatial_filter.weights

        if self.components_to_plot is not None:
            plotter_1 = PlotSpatialFilterComponents(spatial_filter,
                                                    plot_x_lim=self.plot_x_lim,
                                                    plot_y_lim=self.plot_y_lim,
                                                    user_naming_rule=self.user_naming_rule,
                                                    fig_format=self.fig_format,
                                                    components_to_plot=self.components_to_plot,
                                                    domain=self.projection_domain,
                                                    return_figures=self.return_figures,
                                                    save_to_file=self.save_to_file,
                                                    test_frequencies=self.test_frequencies
                                                    )
            plotter_1.run()
            if self.return_figures:
                figures.append(plotter_1.figures)
            if self.plot_projections and self.save_to_file:
                plotter_2 = ProjectSpatialComponents(spatial_filter,
                                                     user_naming_rule=self.user_naming_rule,
                                                     plot_x_lim=self.plot_x_lim,
                                                     plot_y_lim=self.plot_y_lim,
                                                     components_to_plot=self.components_to_plot,
                                                     domain=self.projection_domain,
                                                     test_frequencies=self.test_frequencies,
                                                     return_figures=self.return_figures
                                                     )
                plotter_2.run()
                if self.return_figures:
                    figures += plotter_2.figures
        if self.plot_power:
            _fig_power = self.plot_component_power()
            if self.return_figures:
                figures.append(_fig_power)

        self.figures = figures

        dss_data = []
        p_ratio = spatial_filter.pwr1 / spatial_filter.pwr0

        for _component in np.arange(0, spatial_filter.pwr0.size).astype(int):
            _dss_data = DSSData(
                bias_frequencies=str(self.sf_join_frequencies) if self.sf_join_frequencies is not None else None,
                bias_domain=self.projection_domain,
                unbiased_power=spatial_filter.pwr0[_component],
                biased_power=spatial_filter.pwr1[_component],
                total_unbiased_power=np.sum(spatial_filter.pwr0),
                total_biased_power=np.sum(spatial_filter.pwr1),
                n_components_rotation_1=spatial_filter.n_components_rotation_1,
                n_components_rotation_2=spatial_filter.n_components_rotation_2,
                threshold=self.sf_thr,
                component_rank=_component,
                power_ratio=p_ratio[_component] / np.sum(p_ratio),
                kept=bool(np.isin(_component, filtered_data.kept_components)),
                n_possible_components=spatial_filter.n_possible_components,
                main_channel=None)
            dss_data.append(_dss_data)
        dss_table = pd.DataFrame([_data.__dict__ for _data in dss_data])
        self.output_node.processing_tables_local = Tables(table_name=FilterType.dss,
                                                          data=dss_table,
                                                          data_source=self.name)

    def plot_component_power(self):
        inch = 2.54
        fig = plt.figure()
        fig.set_size_inches(18 / inch, 24 / inch)
        gs = gridspec.GridSpec(3, 1)
        ax = plt.subplot(gs[0, 0])
        _normalized_raw_power = self.pwr0 / np.sum(self.pwr0)
        _normalized_evoked_power = self.pwr1 / np.sum(self.pwr1)

        ax.plot(_normalized_raw_power, marker='o', label='Raw')
        ax.plot(_normalized_evoked_power, marker='o', label='Evoked')
        ax.plot(self.kept_components,
                _normalized_evoked_power[self.kept_components],
                marker='*', label='Kept', linestyle='None')
        ax.axvline(self.kept_components.max(), color='black', label=None)
        ax.set_xlabel('Component rank')
        ax.legend(loc="upper right")
        ax.set_title('Normalized Power')

        ax = plt.subplot(gs[1, 0])

        _accumulated_raw_power = np.cumsum(self.pwr0 / np.sum(self.pwr0))
        _accumulated_evoked_power = np.cumsum(self.pwr1) / np.sum(self.pwr1)
        ax.plot(_accumulated_raw_power, marker='o', label='Raw')
        ax.plot(_accumulated_evoked_power, marker='o', label='Evoked')
        ax.plot(self.kept_components,
                _accumulated_evoked_power[self.kept_components],
                marker='*', label='Kept', linestyle='None')
        ax.axvline(self.kept_components.max(), color='black', label=None)
        ax.set_ylim([0, 1])
        ax.set_xlabel('Component rank')
        ax.set_title('Accumulated power')
        ax.legend(loc="upper right")

        ax = plt.subplot(gs[2, 0])
        _per_component_explained_power = (self.pwr1 / self.pwr0) / np.sum(self.pwr1 / self.pwr0)
        _accumulated_component_explained_power = (np.cumsum(self.pwr1) / np.cumsum(self.pwr0)) / np.sum(
            self.pwr1 / self.pwr0)
        ax.plot(_per_component_explained_power,
                marker='o',
                label='Per component')
        ax.plot(_accumulated_component_explained_power,
                marker='o',
                label='Accumulated')
        ax.plot(self.kept_components,
                _per_component_explained_power[self.kept_components],
                marker='*', label='Kept', linestyle='None')
        ax.axvline(self.kept_components.max(), color='black', label=None)
        # ax.set_ylim([0, 1])
        ax.set_xlabel('Component rank')
        ax.set_title('Evoked power per component / Total power')
        ax.legend(loc="upper right")

        _sep = '_' if self.user_naming_rule is not None else ''
        _naming_rule = '' if self.user_naming_rule is None else self.user_naming_rule
        _figure_basename = self.name + _sep + _naming_rule
        _fig_path = self.output_node.paths.figures_current_dir + _figure_basename + 'components_power' + self.fig_format
        fig.tight_layout()
        fig.savefig(_fig_path)
        print('figure saved: {:}'.format(_fig_path))
        return fig


class PlotSpatialFilterComponents(InputOutputProcess):
    def __init__(self, input_process=SpatialFilter,
                 components_to_plot: np.array = np.arange(0, 10),
                 domain: Domain = Domain.time,
                 test_frequencies: type(u.Quantity) | None = None,
                 user_naming_rule: str = 'components',
                 plot_x_lim: [float, float] = None,
                 plot_y_lim: [float, float] = None,
                 fig_format: str = '.png',
                 return_figures: bool = False,
                 save_to_file: bool = True,

                 **kwargs):
        """
        This class plots the components waveforms (from an SpatialFilter)
        :param input_process: anInputOutputProcess Class
        :param components_to_plot: numpy array indicating which components will be plotted.
        :param domain: indicate the domain in which component will be projected
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param fig_format: string indicating the format of the output figure (e.g. '.png' or '.pdf')
        :param return_figures: bool indicating if figure should be returned in self.figures
        :param save_to_file: bool indicating whether figure should be saved to file
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(PlotSpatialFilterComponents, self).__init__(input_process=input_process, **kwargs)
        self.components_to_plot = components_to_plot
        self.domain = domain
        self.test_frequencies = test_frequencies
        self.user_naming_rule = user_naming_rule
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.fig_format = fig_format
        self.return_figures = return_figures
        self.save_to_file = save_to_file
        self.__kwargs = kwargs

    def transform_data(self):
        if self.components_to_plot is not None:
            # we copy the input process to overwrite the layout without affecting other pipe processes
            _input_process = copy.deepcopy(self.input_process)

            # _input_process.output_node.y_units = u.def_unit('A.U.')
            # generate a generic layout
            _input_process.output_node.layout = np.array([ChannelItem()
                                                          for _ in range(_input_process.output_node.data.shape[1])])
            for _i, _lay in enumerate(_input_process.output_node.layout):
                _lay.label = str(_i)

            if self.domain == Domain.time:
                average = AverageEpochs(_input_process)

            if self.domain == Domain.frequency:
                average = AverageEpochsFrequencyDomain(_input_process,
                                                       test_frequencies=self.test_frequencies,
                                                       delta_frequency=self.input_process.delta_frequency,
                                                       weight_frequencies=self.test_frequencies
                                                       )
            average.name = 'AveragedSpatialComponents'
            average.run()
            plotter = PlotWaveforms(average,
                                    user_naming_rule='{:}_{:}_components'.format(
                                        self.user_naming_rule,
                                        average.output_node.data.shape[1]),
                                    ch_to_plot=self.components_to_plot,
                                    plot_x_lim=self.plot_x_lim,
                                    plot_y_lim=None,
                                    fig_format=self.fig_format,
                                    return_figures=self.return_figures,
                                    save_to_file=self.save_to_file)
            plotter.run()
            self.figures = plotter.figures


class ProjectSpatialComponents(InputOutputProcess):
    def __init__(self, input_process=SpatialFilter,
                 components_to_plot: np.array = np.arange(0, 10),
                 user_naming_rule: str | None = None,
                 plot_x_lim: [float, float] = None,
                 plot_y_lim: [float, float] = None,
                 domain: Domain = Domain.time,
                 test_frequencies: type(u.Quantity) | None = None,
                 return_figures: bool = False,
                 **kwargs):
        """
        This class will project back to the sensor space each component and plot them in a topographic map.
        :param input_process: an SpatialFilter InputOutputProcess Class
        :param components_to_plot: numpy array indicating which components will be plotted.
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param domain: indicate the domain in which component will be projected
        :param test_frequencies: numpy array with frequencies that will be used to compute statistics
        :param return_figures: if true, figures will be returned in self.figures
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ProjectSpatialComponents, self).__init__(input_process=input_process, **kwargs)
        self.components_to_plot = components_to_plot
        self.user_naming_rule = user_naming_rule
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.domain = domain
        self.test_frequencies = set_default_unit(test_frequencies, u.Hz)
        self.return_figures = return_figures
        self.__kwargs = kwargs

    def transform_data(self):
        figures = []
        for _component in self.components_to_plot:
            if _component > self.input_process.output_node.data.shape[1] - 1:
                continue
            if isinstance(self.input_process, SpatialFilter):
                projected_component = ApplySpatialFilter(input_process=self.input_process,
                                                         sf_components=np.array([_component]))
            if isinstance(self.input_process, SpatialFilterICA):
                projected_component = ApplySpatialFilterICA(input_process=self.input_process,
                                                            sf_components=np.array([_component]))
            print('Projecting individual component {:} to sensor space'.format(_component))
            projected_component.run()
            if self.domain == Domain.time:
                average = AverageEpochs(projected_component)
            if self.domain == Domain.frequency:
                average = AverageEpochsFrequencyDomain(projected_component,
                                                       test_frequencies=self.test_frequencies,
                                                       weight_frequencies=self.input_process.sf_join_frequencies,
                                                       delta_frequency=self.input_process.delta_frequency)
            average.run()
            times = None
            if self.domain == Domain.time:
                max_chan = np.argmax(np.std(average.output_node.data, axis=0))
                cha_max_pow_label = average.output_node.layout[max_chan].label
                # find max and min within plot_x_lim
                if self.plot_x_lim is not None:
                    self.plot_x_lim = set_default_unit(self.plot_x_lim, average.output_node.x.unit)
                    _samples = average.output_node.x_to_samples(self.plot_x_lim)
                    _max = np.argmax(average.output_node.data[_samples[0]: _samples[-1], max_chan]) + _samples[0]
                    _min = np.argmin(average.output_node.data[_samples[0]: _samples[-1], max_chan]) + _samples[0]
                else:
                    _max = np.argmax(average.output_node.data[:, max_chan])
                    _min = np.argmin(average.output_node.data[:, max_chan])
                times = np.sort(average.output_node.x[np.array([_max, _min])])
            if self.domain == Domain.frequency:
                if self.input_process.sf_join_frequencies is not None:
                    _idx_f = average.output_node.x_to_samples(self.input_process.sf_join_frequencies)
                    max_chan = np.argmax(np.max(np.abs(average.output_node.data[_idx_f, :]), axis=0))
                    cha_max_pow_label = average.output_node.layout[max_chan].label
                else:
                    _idx_ch = np.argmax(average.output_node.snr)
                    cha_max_pow_label = average.output_node.layout[_idx_ch].label

            # plot time potential fields for average channel
            _sep = '_' if self.user_naming_rule is not None else ''
            _naming_rule = '' if self.user_naming_rule is None else self.user_naming_rule
            plotter = PlotTopographicMap(average,
                                         user_naming_rule=_naming_rule + _sep + 'component_{:}'.format(
                                             _component),
                                         topographic_channels=np.array([cha_max_pow_label]),
                                         plot_x_lim=self.plot_x_lim,
                                         plot_y_lim=self.plot_y_lim,
                                         return_figures=self.return_figures,
                                         title='Component {:}'.format(_component),
                                         times=times)
            plotter.run()
            if self.return_figures and plotter.figures is not None:
                figures += plotter.figures
        self.figures = figures


class CreateAndApplyICAFilter(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 sf_components=np.array([]),
                 sf_thr=0.8,
                 components_to_plot: np.array = np.arange(0, 10),
                 plot_x_lim: [float, float] = None,
                 plot_y_lim: [float, float] = None,
                 n_tracked_points=None,
                 block_size=5,
                 weight_data=True,
                 user_naming_rule: str = '',
                 fig_format: str = '.png',
                 return_figures: bool = False,
                 **kwargs):
        """
        This class will create and apply an ICA filter to the data.
        :param input_process: an InputOutputProcess Class
        :param sf_components: numpy array of integers with indexes of components to be kept
        :param sf_thr: float indicating the percentage of explained variance by components to be kept (this is used when
        sf_components is empty)
        :param components_to_plot: numpy array indicating which components will be plotted.
        :param plot_x_lim: list with minimum and maximum horizontal range of x axis
        :param plot_y_lim: list with minimum and maximum vertical range of y axis
        :param user_naming_rule: string indicating a user naming to be included in the figure file name
        :param fig_format: string indicating the format of the output figure (e.g. '.png' or '.pdf')
        :param return_figures: if True, figures handles will be returnted on self.figures
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(CreateAndApplyICAFilter, self).__init__(input_process=input_process, **kwargs)
        self.sf_components = sf_components
        self.sf_thr = sf_thr
        self.components_to_plot = components_to_plot
        self.plot_x_lim = plot_x_lim
        self.plot_y_lim = plot_y_lim
        self.user_naming_rule = user_naming_rule
        self.fig_format = fig_format
        self.return_figures = return_figures
        self.n_tracked_points = n_tracked_points
        self.block_size = block_size
        self.weight_data = weight_data
        self._unmixing = None
        self._mixing: type(np.array) | None = None
        self._pwr: type(np.array) | None = None
        self._components: type(np.array) | None = None
        self._whitening_m: type(np.array) | None = None
        self.__kwargs = kwargs

    def transform_data(self):
        spatial_filter = SpatialFilterICA(input_process=self.input_process)
        spatial_filter.run()
        print('applying ICA filter')
        filtered_data = ApplySpatialFilterICA(input_process=spatial_filter,
                                              sf_components=self.sf_components,
                                              sf_thr=self.sf_thr,
                                              **self.__kwargs)
        filtered_data.run()
        self.output_node = filtered_data.output_node

        if self.components_to_plot is not None:
            plotter = PlotSpatialFilterComponents(spatial_filter,
                                                  plot_x_lim=self.plot_x_lim,
                                                  plot_y_lim=self.plot_y_lim,
                                                  user_naming_rule=self.user_naming_rule,
                                                  fig_format=self.fig_format,
                                                  components_to_plot=self.components_to_plot
                                                  )
            plotter.run()

            plotter = ProjectSpatialComponents(spatial_filter,
                                               user_naming_rule=self.user_naming_rule,
                                               plot_x_lim=self.plot_x_lim,
                                               plot_y_lim=self.plot_y_lim,
                                               components_to_plot=self.components_to_plot
                                               )
            plotter.run()


class SpatialFilterICA(InputOutputProcess):
    def __init__(self, input_process=InputOutputProcess,
                 **kwargs):
        """
        This class will create an spatial filter based on the data.
        :param input_process:  InputOutputProcess Class
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(SpatialFilterICA, self).__init__(input_process=input_process, **kwargs)
        self.mixing = None
        self.unmixing = None
        self.pwr = None
        self.whitening_m = None
        self.epoch_size = None

    def transform_data(self):
        # compute spatial filter
        _data = self.input_node.data
        self.epoch_size = _data.shape[0]
        if self.input_node.data.ndim == 3:
            _data = et_unfold(self.input_node.data)
        components, unmixing, mixing, pwr, whitening_m = et_ica_epochs(data=_data,
                                                                       tol=1e-4,
                                                                       iterations=10)
        self.mixing = mixing
        self.unmixing = unmixing
        self.pwr = pwr
        self.whitening_m = whitening_m
        self.output_node.data = et_fold(components, epoch_size=self.epoch_size)


class ApplySpatialFilterICA(InputOutputProcess):
    def __init__(self, input_process: SpatialFilterICA | None = None,
                 sf_components: np.array = np.array([]),
                 sf_thr: float = 0.8,
                 **kwargs):
        """
        This class will apply an SpatialFilter to the data. If sf_components are paased, only those will be used to
        filter the data
        :param input_process: an SpatialFilter InputOutputProcess Class
        :param sf_components: numpy array of integers with indexes of components to be kept
        :param sf_thr: float indicating the percentage of explained variance by components to be kept (this is used when
        sf_components is empty)
        :param kwargs: extra parameters to be passed to the superclass
        """
        super(ApplySpatialFilterICA, self).__init__(input_process=input_process, **kwargs)
        self.sf_components = sf_components
        self.sf_thr = sf_thr

    def transform_data(self):
        print('applying spatial filter')
        if self.sf_components is None or not self.sf_components.size:
            cumpower = np.cumsum(self.input_process.pwr) / np.sum(self.input_process.pwr)
            n_idxs = np.argwhere(cumpower <= self.sf_thr)
            n_components = np.arange(n_idxs.size)
        else:
            n_components = self.sf_components

        clean_epochs = self.mix_components(components_idx=n_components)
        if self.input_process.epoch_size is not None:
            clean_epochs = et_fold(clean_epochs, epoch_size=self.input_process.epoch_size)
        self.output_node.data = clean_epochs

    def mix_components(self,
                       components_idx: type(np.array) | None = None):
        w_c = np.zeros(self.input_process.mixing.shape)
        w_c[:, components_idx] = 1
        s_clean = self.input_process.mixing * w_c
        _components = et_unfold(self.input_process.output_node.data)
        clean_data = s_clean.dot(_components.T).T
        return clean_data



import numpy as np
from scipy.stats import f
from scipy import signal
import matplotlib.pyplot as plt
import pyqtgraph as pg
import pyqtgraph.exporters
import pyfftw
import sys
import multiprocessing
from peegy.definitions.eeg_definitions import HotellingTSquareFrequencyTest
from peegy.tools.units.unit_tools import set_default_unit
import astropy.units as u

"""
Created on Mon Dec 15 13:19:14 2014
@author: Jaime Undurraga
"""


class JAverager(object):
    def __init__(self):
        # public
        self.fs = 0
        self.time_vector = np.array([])
        self.time_offset = 0.0
        self.blanking = np.array([])
        self.rejection_window = np.array([])
        self.rejection_level = float("inf")
        self.min_block_size = 5
        self.t_p_snr = np.array([])  # time in indicating the time position of the tracked points to estimate the RN
        self.alpha_level = 0.05  # significance level to do averaging adaptively
        self.filter_window = np.array([])
        self.rejection_window = np.array([])
        self.plot_sweeps = False
        self.splits = 1
        self.split_sweep_count = np.array([])
        self.split_rejected_count = np.array([])
        self.total_sweeps = 0
        self.tracked_points = []
        self.n_noise_sources = []
        self.n_samples_noise_source = []
        self.noise_var = []
        self.min_block_samples = 16
        self.n_samples_theo_rs = 32
        self.fft_analysis = True
        self.n_fft = None
        self.frequencies_to_analyze = np.array([])
        self.high_pass = None
        self.low_pass = None
        self.filt_a = None
        self.filt_b = None
        self.figure_handle = None
        self.lines_per_plot = 20
        self.plot_frequency_range = np.array([])
        self.time_units = u.s
        self.amplitude_units = u.V
        self.demean = True
        self.axis = None
        self.channels_offset = 20.0
        self._time_curves_s = []
        self._time_curves_w = []
        self._freq_curves_s = []
        self._freq_curves_w = []
        self._freq_curves_s_f_bins = []
        self._freq_curves_w_f_bins = []
        self.text_time_item = []
        self.text_freq_item = []
        # private
        self._time = np.array([])
        self._trackedPointsPerSource = np.array([])
        self._filter_hp = np.array([])
        self._filter_lp = np.array([])
        self._self_win_size = 0
        self.__w_average = np.array([])
        self.__s_average = np.array([])
        self._fft = np.array([], dtype=np.complex128)
        self._fft_frequencies = np.array([])
        self._frequency_bin = np.array([], dtype=int)
        self._tracked_frequency_bin = []
        self._block_ave = np.array([])
        self._ave_under_noise_source = np.array([])
        self._block_ave_sweep_count = np.array([])
        self._ave_under_noise_source_sweep_count = np.array([])
        self._split_index = 0
        self._s_snr = np.array([])
        self._win_size = np.array([])
        self._vector_analysis = np.array([])
        self._var_signal_noise_per_n_source = []
        self._analysis_window = np.array([])
        self._current_split_id = None
        self.__sweep_plot_count = 0
        self.__split_plot_count = 0
        self.y_unit = u.uV

    def get_analysis_window(self):
        return self._analysis_window

    def set_analysis_window(self, value):
        self._analysis_window = value

    analysis_window = property(get_analysis_window, set_analysis_window)

    @property
    def time(self):
        return self._time

    @property
    def _w_average(self):
        try:
            value = np.zeros((self._win_size, self.splits)) * self.y_unit
            for i in range(self.splits):
                _remain_noise_var = self.remaining_noise_var
                _noise_var = self.noise_var[i]
                # we make the variance = to 1 if there is no noise to avoid
                # to divide by zero
                _noise_var[_noise_var == 0.0] = 1.0 * self.y_unit ** 2.0
                _remain_noise_var[_remain_noise_var == 0.0] == 1.0 * self.y_unit ** 2.0

                _w_ave = self.__w_average[:, i]
                if self._block_ave_sweep_count[i] < self.min_block_size:
                    value[:, i] = (_w_ave + self._ave_under_noise_source[:, i] *
                                   self.n_samples_noise_source[i][-1] / _noise_var[-1]) / \
                                  (np.sum(self.n_samples_noise_source[i] / _noise_var))
                else:
                    value[:, i] = (_w_ave + self._ave_under_noise_source[:, i] *
                                   self.n_samples_noise_source[i][-1] / _noise_var[-1] + self._block_ave[:, i] *
                                   self._block_ave_sweep_count[i] / _remain_noise_var[i]) / \
                                  (np.sum(np.concatenate((self.n_samples_noise_source[i] / _noise_var,
                                                          np.atleast_1d(
                                                              self._block_ave_sweep_count[i] / _remain_noise_var[i])))))
            return value
        except Exception as e:
            print(_w_ave.shape)
            print(str(e))

    @property
    def w_average(self):
        return self._w_average

    @property
    def w_snr_db(self):
        value = np.zeros(self.splits)
        _snr = self.w_snr
        for i in range(self.splits):
            if _snr[i] == 0.0:
                value[i] = -np.inf * u.dimensionless_unscaled
            else:
                value[i] = 10.0 * np.log10(_snr[i])
        return value

    @property
    def w_snr(self):
        value = np.zeros(self.splits)
        var_sig_plus_noise = np.var(self._w_average[self._vector_analysis, :], ddof=1, axis=0)
        var_noise = self._w_rn ** 2.0
        for i in range(self.splits):
            if var_noise[i] == 0.0:
                value[i] = np.inf * u.dimensionless_unscaled
            else:
                value[i] = np.maximum(var_sig_plus_noise[i] / var_noise[i] - 1.0, 0.0)
        return value

    @property
    def w_signal_variance(self):
        return np.var(self._w_average[self._vector_analysis, :], ddof=1, axis=0)

    @property
    def _w_rn(self):
        return self.w_total_noise_var ** 0.5

    @property
    def w_rn(self):
        return self._w_rn

    @property
    def w_total_noise_var(self):
        value = np.zeros(self.splits) * self.y_unit ** 2.0
        _noise_var_source = self.noise_var_sources
        _n_total_samples_noise_source = self._n_total_samples_noise_source
        for i in range(self.splits):
            with np.errstate(divide='ignore'):
                value[i] = 1.0 / np.sum(_n_total_samples_noise_source[i] / _noise_var_source[i])
        return value

    @property
    def w_total_noise_var_all_splits(self):
        total_noise_var = self._w_rn ** 2.0
        return 1.0 / sum(1./total_noise_var)

    @property
    def w_rn_all_splits(self):
        return self.w_total_noise_var_all_splits ** 0.5

    @property
    def w_snr_all_splits(self):
        var_noise = self.w_total_noise_var_all_splits
        var_sig_plus_noise = np.var(self._w_average_all_splits[self._vector_analysis], ddof=1, axis=0)
        return np.maximum(var_sig_plus_noise / var_noise - 1.0, 0.0)

    @property
    def s_total_noise_var_all_splits(self):
        total_noise_var = self._s_rn ** 2.0
        return 1.0 / sum(1./total_noise_var)

    @property
    def s_rn_all_splits(self):
        return self.s_total_noise_var_all_splits ** 0.5

    @property
    def s_snr_all_splits(self):
        var_noise = self.s_total_noise_var_all_splits
        var_sig_plus_noise = np.var(self._s_average_all_splits[self._vector_analysis], ddof=1, axis=0)
        return np.maximum(var_sig_plus_noise / var_noise - 1.0, 0.0)

    @property
    def _n_total_samples_noise_source(self):
        value = []
        for i in range(self.splits):
            if self._block_ave_sweep_count[i] < self.min_block_samples:
                value.append(self.n_samples_noise_source[i])
            else:
                value.append(np.append(self.n_samples_noise_source[i], self._block_ave_sweep_count[i]))
        return value

    @property
    def noise_var_sources(self):
        _remaining_noise_var = self.remaining_noise_var
        value = []
        for i in range(self.splits):
            if self._block_ave_sweep_count[i] < self.min_block_samples:
                value.append(self.noise_var[i])
            else:
                value.append(np.append(self.noise_var[i], _remaining_noise_var[i]))
        return value

    @property
    def weights(self):
        value = []
        for i in range(self.splits):
            _n_samples = self.n_samples_noise_source[i]
            _var = self.noise_var_sources[i]
            assert _var.size == _n_samples.size
            _w = np.array([])
            for _v, _n in zip(_var, _n_samples):
                _w = np.concatenate((_w, 1 / np.array([_v.value] * int(_n))))
            value.append(_w * 1 / self.y_unit ** 2.0)
        return value

    @property
    def remaining_noise_var(self):
        value = np.zeros(self.splits) * self.y_unit ** 2.0
        for i in range(self.splits):
            n_blocks = int(np.floor(self.split_sweep_count[i] / self.min_block_size))
            ini_pos = n_blocks * self.min_block_size
            if ini_pos < self.tracked_points[i].shape[0]:
                _data = self.tracked_points[i][ini_pos:, :]
            else:
                _data = np.array([])
            if _data.shape[0] <= 1:
                value[i] = 0.0 * self.y_unit ** 2.0
            else:
                value[i] = np.mean(np.var(_data, ddof=1, axis=0))
        return value
    # standard averaging

    @property
    def s_average(self):
        return self.__s_average

    @property
    def s_snr(self):
        value = np.zeros(self.splits)
        var_sig_plus_noise = np.var(self.__s_average[self._vector_analysis, :], ddof=1, axis=0)
        var_noise = self._s_rn ** 2.0
        for i in range(self.splits):
            value[i] = np.maximum(var_sig_plus_noise[i] / var_noise[i] - 1.0, 0.0)
        return value

    @property
    def s_signal_variance(self):
        return np.var(self.__s_average[self._vector_analysis, :], ddof=1, axis=0)

    @property
    def s_snr_db(self):
        value = np.zeros(self.splits)
        _snr = self.s_snr
        for i in range(self.splits):
            if _snr[i] == 0.0:
                value[i] = -np.inf * u.dimensionless_unscaled
            else:
                value[i] = 10.0 * np.log10(_snr[i])
        return value

    @property
    def _s_rn(self):
        _noise_var_source = self.noise_var_sources
        value = np.zeros(self.splits) * self.y_unit ** 2.0
        _n_total_samples_noise_source = self._n_total_samples_noise_source
        for i in range(self.splits):
            value[i] = np.sum(_noise_var_source[i] * _n_total_samples_noise_source[i]) / (np.sum(
                _n_total_samples_noise_source[i]) ** 2.0)
        return value ** 0.5

    @property
    def s_rn(self):
        return self._s_rn

    @property
    def w_fft(self):
        fft = pyfftw.builders.rfft(self.w_average, n=self.n_fft, overwrite_input=False, planner_effort='FFTW_ESTIMATE',
                                   axis=0, threads=multiprocessing.cpu_count())
        return fft() * self.y_unit

    @property
    def s_fft(self):
        fft = pyfftw.builders.rfft(self.s_average, n=self.n_fft, overwrite_input=False, planner_effort='FFTW_ESTIMATE',
                                   axis=0, threads=multiprocessing.cpu_count())
        return fft() * self.y_unit

    @property
    def s_fft_spectral_ave_all_splits(self):
        _s_snr = self.s_snr
        if np.sum(_s_snr):
            out = np.sum(np.abs(np.fft.rfft(self.s_average, axis=0)) * _s_snr.T, axis=1) / np.sum(_s_snr)
        else:
            out = np.mean(np.abs(np.fft.rfft(self.s_average, axis=0)), axis=1)
        return out

    @property
    def w_fft_spectral_ave_all_splits(self):
        _w_snr = self.w_snr
        return np.sum(np.abs(np.fft.rfft(self.w_average, axis=0)) * _w_snr.T, axis=1) / np.sum(_w_snr)

    def w_fft_spectral_weighted_ave_all_splits(self, frequency_index=int(0)):
        _h_test = self.hotelling_t_square_test()
        _snr = np.array([x.snr for _split in _h_test for _i, x in enumerate(_split) if _i == frequency_index])
        if np.all(_snr == 0):
            _snr[:] = 1.0
        return np.sum(np.abs(np.fft.rfft(self.w_average, axis=0)) * _snr.T, axis=1) / np.sum(_snr)

    def w_average_spectral_weighted_ave_all_splits(self, frequency_index=int(0)):
        _h_test = self.hotelling_t_square_test()
        _snr = np.array([x.snr for _split in _h_test for _i, x in enumerate(_split) if _i == frequency_index])
        if np.all(_snr == 0):
            _snr[:] = 1.0
        return np.sum(self.w_average * _snr.T, axis=1) / np.sum(_snr)

    def samples_to_time(self, in_samples):
        return np.array(in_samples).astype(np.float64) / self.fs - self.time_offset

    def time_to_samples(self, in_time):
        return np.rint((in_time + self.time_offset) * self.fs).astype(np.int64)

    def add_sweep(self, new_sweep=np.array([]), split_id=None):
        try:
            # initialization of parameters
            if self.total_sweeps == 0:
                self.y_unit = set_default_unit(new_sweep, u.uV).unit
                self._win_size = new_sweep.size
                self._time = np.arange(self._win_size) / self.fs
                ms = self._win_size
                _t_snr = self.time_to_samples(self.t_p_snr)
                self._s_snr = _t_snr[_t_snr <= self._win_size - 1]

                if not np.alltrue(_t_snr < ms):
                    print('Some tracking points were omitted')

                if self.analysis_window.size == 0:
                    self.analysis_window = self.samples_to_time([0, ms - 1])
                    self._vector_analysis = list(range(0, ms - 1))
                else:
                    self.analysis_window[0] = max(self.analysis_window[0], 0)
                    self.analysis_window[1] = min(self.analysis_window[1], self.samples_to_time(ms - 1))
                    _samples = self.time_to_samples(self.analysis_window)
                    self._vector_analysis = list(range(_samples[0], _samples[1] + 1))

                if self.filter_window.size == 0:
                    self.filter_window = self.samples_to_time([0, ms - 1])

                if self.rejection_window.size == 0:
                    self.rejection_window = self.samples_to_time([0, ms - 1])

                self.time_vector = self.samples_to_time(list(range(ms))) - self.time_offset
                self.__w_average = np.zeros((ms, self.splits)) * 1 / self.y_unit  # values are divided by variance
                self.__s_average = np.zeros((ms, self.splits)) * self.y_unit
                # self._WSAverage = np.zeros(m`s,self.splits);

                if self.n_fft is None:
                    self.n_fft = ms

                if np.mod(self.n_fft, 2) == 0:
                    _fft_size = int((self.n_fft / 2) + 1)
                else:
                    _fft_size = int((self.n_fft + 1) / 2)

                self._fft = np.zeros((_fft_size, self.splits), dtype=np.complex) * self.y_unit
                self._fft_frequencies = np.arange(_fft_size) * self.fs / self.n_fft
                self._block_ave = np.zeros((ms, self.splits)) * self.y_unit
                self._ave_under_noise_source = np.zeros((ms, self.splits)) * self.y_unit
                self.split_sweep_count = np.zeros(self.splits)
                self.split_rejected_count = np.zeros(self.splits)
                self._block_ave_sweep_count = np.zeros(self.splits)
                self._ave_under_noise_source_sweep_count = np.zeros(self.splits)

                for j in range(self.splits):
                    self.tracked_points.append([])
                    self._tracked_frequency_bin.append([])
                    self.n_noise_sources.append([1])
                    self.n_samples_noise_source.append(np.array([0.0]))
                    self.noise_var.append(np.array([0.0]) * self.y_unit ** 2.0)
                    self._var_signal_noise_per_n_source.append(np.array([0.0]) * self.y_unit ** 2.0)
                    # self._TrackedFreqBin[i] = []
                    # setFilter(self)

            if new_sweep.size != self._win_size:
                raise AssertionError

            # index of split
            if split_id is not None:
                self._split_index = np.minimum(split_id, self.splits - 1)
            else:
                self._split_index = np.mod(self.total_sweeps, self.splits)

            # blanking
            # raw_sweep = new_sweep;
            # new_sweep = doBlanking(self,new_sweep);
            #
            if self.demean:
                new_sweep = new_sweep - np.mean(new_sweep, axis=0)
            # rejection
            if self.do_rejection(new_sweep):
                if self.plot_sweeps:
                    self.plot_steps(new_sweep, new_sweep, True)
                return

            # filtering
            new_sweep = self.do_filtering(new_sweep)

            # standard averaging
            self._do_s_average(new_sweep)
            # weighted average
            self._add_tracked_point(new_sweep)
            # do fft
            if self.fft_analysis:
                self._do_frequency_average(new_sweep)

            # weighted average using each sweepVariance
            self._do_w_average(new_sweep)
            # plotSteps2(self,new_sweep,raw_sweep,false)
            # update index
            self.split_sweep_count[self._split_index] += 1
            # plot steps
            if self.plot_sweeps:
                self.plot_steps(new_sweep, False)
            # updating total and current index
            self.total_sweeps = self.total_sweeps + 1
            print('channel {:f}, sweeps {:f}'.format(self._split_index, self.split_sweep_count[self._split_index]))

        except Exception as e:
            print(str(e))

    def do_filtering(self, data=np.array([])):
        filtered_data = data
        if self.filt_a is None and self.filt_b is None:
            if self.low_pass not in [None, 0] and self.high_pass in [None, 0]:
                low = 2 * self.low_pass / self.fs
                self.filt_b, self.filt_a = signal.butter(N=3, Wn=low, btype='lowpass', analog=False)
            if self.high_pass not in [None, 0] and self.low_pass in [None, 0]:
                high = 2 * self.high_pass / self.fs
                self.filt_b, self.filt_a = signal.butter(N=3, Wn=high, btype='highpass', analog=False)
            if self.high_pass and self.low_pass:
                low = 2 * self.low_pass / self.fs
                high = 2 * self.high_pass / self.fs
                self.filt_b, self.filt_a = signal.butter(N=3, Wn=[high, low], btype='bandpass', analog=False)
        if self.filt_a is not None and self.filt_b is not None:
            flip_data = np.flip(data, axis=0)
            filtered_data = np.concatenate((flip_data[:-1], data, flip_data[1::]))
            filtered_data = signal.filtfilt(self.filt_b, self.filt_a, filtered_data, axis=0)
            filtered_data = filtered_data[flip_data.size - 1: flip_data.size + data.size - 1]
        return filtered_data

    def _do_s_average(self, new_sweep):
        if self.split_sweep_count[self._split_index] == 0:
            self.__s_average[:, self._split_index] = new_sweep
        else:
            self.__s_average[:, self._split_index] = (self.__s_average[:, self._split_index] *
                                                      self.split_sweep_count[self._split_index] + new_sweep) / \
                                                     (self.split_sweep_count[self._split_index] + 1)

    def _do_frequency_average(self, new_sweep):
        fft = pyfftw.builders.rfft(new_sweep, n=self.n_fft, overwrite_input=False, planner_effort='FFTW_ESTIMATE',
                                   axis=0, threads=multiprocessing.cpu_count())
        _fft = fft() * self.y_unit
        if self.total_sweeps == 0:
            if self.frequencies_to_analyze is not None:
                for fin in self.frequencies_to_analyze:
                    self._frequency_bin = np.append(self._frequency_bin, np.argmin(
                        np.abs(self._fft_frequencies - fin)))
        if self.split_sweep_count[self._split_index] == 0:
            self._fft[:, self._split_index] = _fft
            self._tracked_frequency_bin[self._split_index].append(_fft[self._frequency_bin].value)
        else:
            self._fft[:, self._split_index] = (self._fft[:, self._split_index] *
                                               self.split_sweep_count[self._split_index] + _fft) / \
                                              (self.split_sweep_count[self._split_index] + 1)
            self._tracked_frequency_bin[self._split_index] = np.vstack(
                (self._tracked_frequency_bin[self._split_index],
                 _fft[self._frequency_bin].value))

    def _add_tracked_point(self, new_sweep):
        if self.split_sweep_count[self._split_index] == 0:
            self.tracked_points[self._split_index] = new_sweep[self._s_snr][None, :]
        else:
            self.tracked_points[self._split_index] = np.vstack((self.tracked_points[self._split_index],
                                                                new_sweep[self._s_snr]))

    def _add_ave_under_noise_source(self):
        self._ave_under_noise_source[:, self._split_index] = \
            (self._ave_under_noise_source_sweep_count[self._split_index] *
             self._ave_under_noise_source[:, self._split_index] +
             self._block_ave[:, self._split_index]) / \
            (self._ave_under_noise_source_sweep_count[self._split_index] + 1)

        self._ave_under_noise_source_sweep_count[self._split_index] += 1

    def _do_w_average(self, new_sweep):
        # average of block
        self._block_ave[:, self._split_index] = (self._block_ave_sweep_count[self._split_index] *
                                                 self._block_ave[:, self._split_index] + new_sweep) / \
                                                (self._block_ave_sweep_count[self._split_index] + 1)

        self._block_ave_sweep_count[self._split_index] += 1
        self.n_samples_noise_source[self._split_index][-1] += 1
        # source segmentation
        if np.mod(self.split_sweep_count[self._split_index] + 1, self.min_block_size) == 0:
            n_blocks = int(np.round((self.split_sweep_count[self._split_index] + 1) / self.min_block_size))
            # calculate the mean of the variance of tracked points
            ini_pos = int((n_blocks - 1) * self.min_block_size)
            _noise_var = np.mean(np.var(self.tracked_points[self._split_index][ini_pos:, :], ddof=1, axis=0))
            if n_blocks == 1:
                self.noise_var[self._split_index][-1] = _noise_var
                # self.n_samples_noise_source[self._split_index][-1] = self.min_block_size
                _new_var = np.var(self._block_ave[self._vector_analysis, self._split_index],
                                  ddof=1, axis=0) * self.min_block_size
                self._var_signal_noise_per_n_source[self._split_index][-1] = _new_var
                self._add_ave_under_noise_source()
            else:
                # noise source segmentation
                self.n_samples_noise_source[self._split_index][-1] -= self.min_block_size
                d_old = self.t_p_snr.size * self.n_samples_noise_source[self._split_index][-1] - 1
                d_new = self.t_p_snr.size * self.min_block_size - 1
                f_val = self.noise_var[self._split_index][-1] / _noise_var if _noise_var != 0 else \
                    np.inf * u.dimensionless_unscaled
                p = f.cdf(f_val, d_old, d_new)
                # if is the same source of noise
                if (p > self.alpha_level) & (p < 1 - self.alpha_level):
                    self._add_ave_under_noise_source()

                    q = self.n_samples_noise_source[self._split_index][-1] / self.min_block_size

                    self.noise_var[self._split_index][-1] = (q * self.noise_var[self._split_index][-1] + _noise_var) / \
                                                            (q + 1.0)

                    self.n_samples_noise_source[self._split_index][-1] += self.min_block_size
                    self._var_signal_noise_per_n_source[self._split_index][-1] = \
                        np.var(self._ave_under_noise_source[self._vector_analysis, self._split_index],
                               ddof=1, axis=0) * self.n_samples_noise_source[self._split_index][-1]
                else:
                    _noise_variance = self.noise_var[self._split_index]
                    # here we protect the the average when the variance is zero
                    _noise_variance[_noise_variance == 0.0] = 1.0 * self.y_unit ** 2.0

                    self.__w_average[:, self._split_index] = \
                        self.__w_average[:, self._split_index] + \
                        self._ave_under_noise_source[:, self._split_index] * \
                        self.n_samples_noise_source[self._split_index][-1] / _noise_variance[-1]

                    self._var_signal_noise_per_n_source[self._split_index] = np.append(
                        self._var_signal_noise_per_n_source[self._split_index],
                        np.var(self._ave_under_noise_source[self._vector_analysis, self._split_index], ddof=1, axis=0) *
                        self.min_block_size)

                    self.n_samples_noise_source[self._split_index] = np.append(
                        self.n_samples_noise_source[self._split_index], self.min_block_size)

                    self.noise_var[self._split_index] = np.append(self.noise_var[self._split_index], _noise_var)

                    self._ave_under_noise_source_sweep_count[self._split_index] = 0

                    self._add_ave_under_noise_source()

                    self.n_noise_sources[self._split_index][-1] += 1

                # restarting block averaging
            self._block_ave_sweep_count[self._split_index] = 0

    @property
    def _w_cum_noise_var(self):
        value = []
        for i in range(self.splits):
            value.append(1.0 / np.cumsum(self._n_total_samples_noise_source[i] / self.noise_var_sources[i]))
        return value

    @property
    def _s_cum_noise_var(self):
        value = []
        n_total_samples_noise = self._n_total_samples_noise_source
        noise_var_sources = self.noise_var_sources
        for i in range(self.splits):
            value.append(np.cumsum(noise_var_sources[i] * n_total_samples_noise[i]) / (np.cumsum(
                n_total_samples_noise[i]) ** 2.0))
        return value

    def plot_noise(self):
        scale = 1
        n_total_samples_noise = self._n_total_samples_noise_source
        cum_noise_var_s_ave = self._s_cum_noise_var
        cum_noise_var_w_ave = self._w_cum_noise_var
        # _, ax = plt.subplots(1, 1, sharey=True)
        # if not isinstance(ax, np.ndarray):
        #     ax = np.array([ax])
        for i in range(self.splits):
            ref_ini_noise_rms = np.mean(np.var(self.tracked_points[i][list(range(self.n_samples_theo_rs)), :],
                                               ddof=1, axis=0)) ** 0.5
            num_samples = np.cumsum(n_total_samples_noise[i])
            # get standard and weighted variances
            plt.plot(num_samples, ref_ini_noise_rms / (num_samples ** 0.5) * scale, marker='o', label='Theoretical')
            plt.plot(num_samples, cum_noise_var_s_ave[i] ** 0.5 * scale, marker='s', label='RNs')
            plt.plot(num_samples, cum_noise_var_w_ave[i] ** 0.5 * scale, marker='v', label='RNw')
            plt.legend(loc='upper right')

        plt.xlabel('Averaged sweeps')
        plt.ylabel(r'Residual noise RMS [$\mu$V]')

        plt.show()

    def hotelling_t_square_test(self, alpha=0.05):
        """
        Computes hotelling t2 on frequency bins.
        :param alpha: statistical level for significant responses
        :return: a list (for each split). Each list contains the statistical results for each frequency in a
        HotellingTSquareFrequencyTest class
        """
        value = []
        try:
            _fft = self.w_fft
            for i in np.arange(self.splits):
                _split_data = np.array([])
                _weights = self.weights[i]
                for j in np.arange(self._frequency_bin.size):
                    rx = np.real(self._tracked_frequency_bin[i][:, j]) * self.y_unit
                    ix = np.imag(self._tracked_frequency_bin[i][:, j]) * self.y_unit
                    # for standard averaging
                    # mean_rx = np.mean(rx)
                    # mean_ix = np.mean(ix)

                    # for weighted averaging
                    mean_rx = np.sum(_weights * rx) / np.sum(_weights)
                    mean_ix = np.sum(_weights * ix) / np.sum(_weights)
                    mean_xy = np.array([mean_rx.value, mean_ix.value])[None, :] * self.y_unit
                    _t_square = np.inf * u.dimensionless_unscaled

                    # for standard averaging
                    # dof = self.split_sweep_count[i]

                    # for weighted averaging
                    dof = np.sum(_weights) ** 2.0 / np.sum(_weights ** 2.0)

                    with np.errstate(divide='ignore', invalid='ignore'):
                        if np.linalg.cond(np.vstack((rx, ix))) < 1 / sys.float_info.epsilon:
                            x = np.vstack((rx, ix))
                            # for standard averaging
                            # _cov_mat = np.cov(x)

                            # for weighted averaging
                            _cov_mat = np.cov(x.value, aweights=_weights.value) * x.unit ** 2.0
                            # _eigen_vals = np.linalg.eigvals(_cov_mat)
                            if np.linalg.cond(_cov_mat) < 1 / sys.float_info.epsilon:
                                # get tsquare
                                _t_square = np.dot(dof * mean_xy,
                                                   np.dot(np.linalg.inv(_cov_mat), mean_xy.T)).squeeze()

                    _f = (dof - 2) / (2 * dof - 2) * _t_square
                    _d1 = 2
                    _d2 = dof - 2
                    # we calculate the half length and direction of the ellipse
                    _f_critic = f.ppf(1 - alpha, _d1, _d2)
                    # compute residual noise from circular standard error
                    _rn = np.sqrt(np.var(rx + 1j*ix, ddof=1) / dof)
                    _amp = np.abs(_fft[self._frequency_bin[j], i])
                    _phase = np.angle(_fft[self._frequency_bin[j], i])
                    # _L = ((2 * (self.split_sweep_count[i] - 1) * D * F95)./(
                    # self.split_sweep_count(i) * (self.split_sweep_count(i) - 2))) ** 2
                    _snr = np.maximum(_amp ** 2.0 / _rn ** 2.0 - 1, 0.0 * u.dimensionless_unscaled)
                    test = HotellingTSquareFrequencyTest(
                        frequency_tested=self._fft_frequencies[self._frequency_bin[j]],
                        t_square=_t_square,
                        df_1=_d1,
                        df_2=_d2,
                        n_epochs=self.split_sweep_count[i],
                        f=_f,
                        p_value=1 - f.cdf(_f, _d1, _d2),
                        amplitude=_amp,
                        phase=_phase,
                        rn=_rn,
                        snr=_snr,
                        snr_db=10 * np.log10(_snr) if _snr > 0.0 else -np.inf * u.dimensionless_unscaled,
                        f_critic=_f_critic,
                        channel=str(i))
                    _split_data = np.append(_split_data, test)
                value.append(_split_data)
        except Exception as e:
            print(str(e))
        return value

    @property
    def _w_average_all_splits(self):
        _w_snr = self.w_snr
        if sum(_w_snr):
            out = np.sum(self._w_average * _w_snr.T, axis=1) / sum(_w_snr)
        else:
            out = np.mean(self._w_average * _w_snr.T, axis=1)
        return out

    @property
    def w_average_all_splits(self):
        return self._w_average_all_splits

    @property
    def _s_average_all_splits(self):
        _s_snr = self.s_snr
        if sum(_s_snr):
            out = np.sum(self.__s_average * _s_snr.T, axis=1) / sum(_s_snr)
        else:
            out = np.mean(self.__s_average * _s_snr.T, axis=1)
        return out

    @property
    def s_average_all_splits(self):
        return self._s_average_all_splits

    def do_rejection(self, new_sweep):
        value = False
        if not self.rejection_window.size:
            return value
        _samples = np.sort(self.time_to_samples(self.rejection_window))
        _pos_ini = np.minimum(_samples[0], self._win_size)
        _pos_end = np.minimum(_samples[1], self._win_size)
        _amp_max = np.max(np.abs(new_sweep[_pos_ini: _pos_end]))
        if _amp_max > self.rejection_level:
            self.split_rejected_count[self._split_index] += 1
            value = True
            print(('Rejected Split-' + str(self._split_index) + ':' +
                  str(self.split_rejected_count[self._split_index])))
        return value

    def plot_steps(self, new_sweep, rejected, generate_video=False):
        if not np.alltrue(self.split_sweep_count > self.min_block_samples):
            return
        raw_sweep = new_sweep
        if np.mod(self.total_sweeps, self.min_block_samples) == 0 or \
                ((self.splits == 1) and np.mod(self.total_sweeps, self.min_block_samples + 1) == 0) or \
                np.mod(self.split_sweep_count[self._split_index] + 1, self.min_block_samples) == 0:
            if not isinstance(self.figure_handle, pg.graphicsWindows.GraphicsWindow):
                win = pg.GraphicsWindow()
                ax = [[None] * 3 for n_item in range(self.splits)]
                for n_row in range(self.splits):
                    for n_col in range(3):
                        ax[n_row][n_col] = win.addPlot(row=n_row, col=n_col)
                        if n_col <= 1:
                            ax[n_row][n_col].setLabel('bottom', "Time [ms]")
                            ax[n_row][n_col].setLabel('left', "Amplitude")
                        if n_col == 2:
                            ax[n_row][n_col].setLabel('bottom', "Frequency [Hz]")
                            ax[n_row][n_col].setLabel('left', "Amplitude")
                            if self.plot_frequency_range.size == 0:
                                _frequency_range = [0, self.fs / 2]
                            else:
                                _frequency_range = np.maximum(np.minimum(self.plot_frequency_range, self.fs / 2), 0)
                            ax[n_row][n_col].setXRange(_frequency_range[0], _frequency_range[1])
                            ax[n_row][n_col].setYRange(-10, 10)
                self.figure_handle = win
            else:
                ax = [[None] * 3 for n_item in range(self.splits)]
                for n_row in range(self.splits):
                    for n_col in range(3):
                        ax[n_row][n_col] = self.figure_handle.getItem(row=n_row, col=n_col)
            _time = self._time * 1000

            if np.mod(self.__sweep_plot_count, self.lines_per_plot) == 0:
                for y in np.arange(self.splits):
                    [ax[y][0].addItem(pg.InfiniteLine(pos=x, angle=90)) for x in self.rejection_window]
                    if self.rejection_level != np.inf:
                        [ax[y][0].addItem(pg.InfiniteLine(pos=x, angle=0)) for x in [self.rejection_level,
                                                                                     -self.rejection_level]]
                self.__sweep_plot_count = 1

            self.__sweep_plot_count += 1

            _clear = np.mod(self.__sweep_plot_count, self.lines_per_plot) == 0
            ax[self._split_index][0].setLabel('top', 'Total sweeps:' + str(self.split_sweep_count[self._split_index] +
                                                                           self.split_rejected_count[self._split_index])
                                              )

            if not rejected:
                ax[self._split_index][0].plot(_time, raw_sweep, pen=(200, 200, 200), clear=_clear)
                ax[self._split_index][0].plot(_time, new_sweep, pen=(255, 255, 255))
                ax[self._split_index][0].plot(self.t_p_snr * 1000, new_sweep[self.time_to_samples(self.t_p_snr)],
                                              symbol='o', symbolPen='b',  pen=None)
                pg.QtCore.QCoreApplication.processEvents()
            else:
                ax[self._split_index][0].plot(_time, raw_sweep, pen=(255, 0, 0), clear=_clear)
                ax[self._split_index][0].plot(_time, new_sweep, pen=(255, 0, 0))
                pg.QtCore.QCoreApplication.processEvents()
                return

            if not np.alltrue(self.split_sweep_count > self.min_block_samples):
                return
            ax[self._split_index][1].plot(_time, self.s_average[:, self._split_index], clear=True, pen=(0, 255, 255))
            ax[self._split_index][1].plot(_time, self.w_average[:, self._split_index], pen=(255, 255, 0))
            ax[self._split_index][1].setLabel('top', 'Accepted sweeps:' +
                                              str(self.split_sweep_count[self._split_index]) + '/' +
                                              'SNRs[dB]:' + '{:.1f}'.format(self.s_snr_db[self._split_index]) + '/' +
                                              'SNRw[dB]:' + '{:.1f}'.format(self.w_snr_db[self._split_index]))

            ax[self._split_index][2].plot(self._fft_frequencies, 2 * np.abs(self.s_fft[:, self._split_index]) /
                                          self._win_size,
                                          clear=True, pen=(0, 255, 255))
            ax[self._split_index][2].plot(self._fft_frequencies, 2 * np.abs(self.w_fft[:, self._split_index]) /
                                          self._win_size, pen=(255, 255, 0))

            # plot frequency bins of interest
            ax[self._split_index][2].plot(self._fft_frequencies[self._frequency_bin], 2 *
                                          np.abs(self.s_fft[self._frequency_bin, self._split_index]) /
                                          self._win_size, symbol='t', symbolPen='b',  pen=None)
            ax[self._split_index][2].plot(self._fft_frequencies[self._frequency_bin], 2 *
                                          np.abs(self.w_fft[self._frequency_bin, self._split_index]) /
                                          self._win_size, symbol='t', symbolPen='b',  pen=None)
            _h_test = self.hotelling_t_square_test()
            if len(_h_test) > self._split_index:
                for _h_t_, _f_i_ in zip(_h_test[self._split_index], self._frequency_bin):
                    text = '{:.1f}'.format(_h_t_.snr)
                    text_item = pg.TextItem(text, anchor=(0.0, 0.9))
                    text_item.setPos(self._fft_frequencies[_f_i_], 2 *
                                     np.abs(self.w_fft[_f_i_, self._split_index]) / self._win_size)
                    ax[self._split_index][2].addItem(text_item)

            pg.QtCore.QCoreApplication.processEvents()
            if generate_video:
                exporter = pg.exporters.ImageExporter(plt.plotItem)
                exporter.parameters()['width'] = 100
                exporter.export('fileName.png')

    def initialize_plot_current_data(self):
        if self.figure_handle is None:
            self.figure_handle = pg.GraphicsWindow()
        n_cols = 1 + self.fft_analysis
        win = self.figure_handle
        self.axis = [None] * n_cols
        self.channels_offset = 20.0
        # for n_row in range(self.splits):
        for n_col in range(n_cols):
            self.axis[n_col] = win.addPlot(row=1, col=n_col)
            self.axis[n_col].addLegend()
            if n_col == 0:
                self.axis[n_col].setLabel('bottom', "Time [ms]")
                self.axis[n_col].setLabel('left', "Amplitude")
                self.axis[n_col].setXRange(self.time[0] * 1000, self.time[-1] * 1000)
                [self.axis[n_col].addItem(pg.InfiniteLine(_pos)) for _pos in self._analysis_window * 1000.]
                dammy_data = np.zeros((self.time.size, self.splits), dtype=float)

                for _i in range(self.splits):
                    name_legend = 'S_ave' if _i == 0 else None
                    self._time_curves_s.append(self.axis[n_col].plot(self.time, dammy_data[:, _i] -
                                                                     _i * self.channels_offset,
                                                                     pen=pg.intColor(0, 2), name=name_legend))

                    text_item = pg.TextItem('', anchor=(-0.1, 0.0))
                    text_item.setPos(0, - _i * self.channels_offset)
                    self.text_time_item.append(text_item)
                    self.axis[n_col].addItem(text_item)
                    name_legend = 'W_ave' if _i == 0 else None
                    self._time_curves_w.append(self.axis[n_col].plot(self.time, dammy_data[:, _i] -
                                                                     _i * self.channels_offset,
                                                                     pen=pg.intColor(1, 2), name=name_legend))
            if n_col == 1:
                for _i in range(self.splits):
                    dammy_data = np.zeros((self._fft_frequencies.size, self.splits), dtype=float)
                    name_legend = 'S_ave' if _i == 0 else None
                    self._freq_curves_s.append(self.axis[n_col].plot(self._fft_frequencies, dammy_data[:, _i] -
                                                                     _i * self.channels_offset,
                                                                     pen=pg.intColor(0, 2), name=name_legend))
                    name_legend = 'W_ave' if _i == 0 else None
                    self._freq_curves_w.append(self.axis[n_col].plot(self._fft_frequencies, dammy_data[:, _i] -
                                                                     _i * self.channels_offset,
                                                                     pen=pg.intColor(1, 2), name=name_legend))

                    self._freq_curves_s_f_bins.append(self.axis[n_col].plot(self._fft_frequencies[self._frequency_bin],
                                                                            dammy_data[self._frequency_bin, _i] -
                                                                            _i * self.channels_offset,
                                                                            symbol='t', symbolPen='b', pen=None))
                    self._freq_curves_w_f_bins.append(self.axis[n_col].plot(self._fft_frequencies[self._frequency_bin],
                                                                            dammy_data[self._frequency_bin, _i] -
                                                                            _i * self.channels_offset,
                                                                            symbol='t', symbolPen='b', pen=None))
                    self.text_freq_item.append([])
                    for _f_i, f_b in enumerate(self._frequency_bin):
                        text_item = pg.TextItem('here', anchor=(-0.1, 0.0))
                        text_item.setPos(self._fft_frequencies[f_b], - _i * self.channels_offset)
                        self.text_freq_item[-1].append(text_item)
                        self.axis[n_col].addItem(text_item)

                self.axis[n_col].setLabel('bottom', "Frequency [Hz]")
                self.axis[n_col].setLabel('left', "Amplitude")
                if self.plot_frequency_range.size == 0:
                    _frequency_range = [0, self.fs / 2]
                else:
                    _frequency_range = np.maximum(np.minimum(self.plot_frequency_range, self.fs / 2), 0)
                self.axis[n_col].setXRange(_frequency_range[0], _frequency_range[1])
            self.axis[n_col].setYRange(-self.channels_offset * self.splits, self.channels_offset)
            self.axis[n_col].showGrid(x=True, y=True, alpha=0.8)

    def plot_current_data(self):
        # try:
        if not self.split_sweep_count.size:
            return
        if self.axis is None:
            self.initialize_plot_current_data()
        _time = self._time * 1000

        if np.mod(self.__sweep_plot_count, self.lines_per_plot) == 0:
            self.__sweep_plot_count = 1
        self.__sweep_plot_count += 1

        for _split_index in range(self.splits):

            if not np.alltrue(self.split_sweep_count > self.min_block_samples):
                return
            self._time_curves_s[_split_index].setData(_time,
                                                      self.s_average[:, _split_index] -
                                                      _split_index * self.channels_offset)

            self._time_curves_w[_split_index].setData(_time,
                                                      self.w_average[:, _split_index] -
                                                      _split_index * self.channels_offset)

            text = '{:.1f} / {:.1f} \n SNRs[dB]: {:.1f} SNRw[dB] {:.1f}'.format(self.split_sweep_count[_split_index],
                                                                                self.split_sweep_count[_split_index] +
                                                                                self.split_rejected_count[_split_index],
                                                                                self.s_snr_db[_split_index],
                                                                                self.w_snr_db[_split_index])
            self.text_time_item[_split_index].setText(text)

            if self.fft_analysis:
                self._freq_curves_s[_split_index].setData(self._fft_frequencies,
                                                          2 * np.abs(self.s_fft[:, _split_index]) / self._win_size -
                                                          _split_index * self.channels_offset)

                self._freq_curves_w[_split_index].setData(self._fft_frequencies,
                                                          2 * np.abs(self.w_fft[:, _split_index]) / self._win_size -
                                                          _split_index * self.channels_offset)

                # plot frequency bins of interest
                self._freq_curves_s_f_bins[_split_index].setData(self._fft_frequencies[self._frequency_bin], 2 *
                                                                 np.abs(self.s_fft[self._frequency_bin, _split_index]) /
                                                                 self._win_size - _split_index * self.channels_offset)

                self._freq_curves_w_f_bins[_split_index].setData(self._fft_frequencies[self._frequency_bin], 2 *
                                                                 np.abs(self.w_fft[self._frequency_bin, _split_index]) /
                                                                 self._win_size - _split_index * self.channels_offset)

                _h_test = self.hotelling_t_square_test()

                if len(_h_test) > _split_index:
                    for _b_i, (_h_t_, _f_i_) in enumerate(zip(_h_test[_split_index], self._frequency_bin)):
                        text = '{:.1f}'.format(_h_t_.snr)
                        self.text_freq_item[_split_index][_b_i].setPos(self._fft_frequencies[_f_i_], 2 *
                                                                       np.abs(self.w_fft[_f_i_, _split_index]) /
                                                                       self._win_size - _split_index *
                                                                       self.channels_offset)
                        self.text_freq_item[_split_index][_b_i].setText(text)

        # pg.QtCore.QCoreApplication.processEvents()
        # except Exception as e:
        #     print str(e)

    @property
    def win_size(self):
        return self._win_size



# -*- coding: utf-8 -*-
"""
Created on Tue Dec 16 10:55:37 2014

@author: jundurraga-ucl
"""
import numpy as np
import matplotlib.pyplot as plt
import pyfftw
# Enable below for interactive backend
# import matplotlib
# if 'qt5agg' in matplotlib.backends.backend_registry.list_builtin():
#     matplotlib.use('qt5agg')

# create synthetic data
fs = 1000.0
nsamples = np.round(1 * fs).astype(int)
nchans = 5
ntrials = 1
noise_dim = 300  # dimensionality of noise
f1 = 40
time = np.arange(nsamples) / fs
source = np.expand_dims(np.sin(2 * np.pi * f1 * time), axis=1)
coeff = np.arange(1, nchans)
coeff = np.expand_dims(np.hstack((coeff, coeff)), 0)

s = source * coeff
s_std = np.std(s, axis=0)
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))

data = s

fft = pyfftw.builders.rfft(
                s,
                overwrite_input=False,
                planner_effort="FFTW_ESTIMATE",
                axis=0,
                threads=1,
                n=s.shape[0],
            )
fft_y = fft()
psd = 2 * np.abs(fft_y) ** 2 / (s.shape[0] ** 2)
w_fft = fft_y * 2 / s.shape[0]
amp_from_psd = (2 * psd) ** 0.5
freq = np.arange(0, w_fft.shape[0]) * fs / s.shape[0]
plt.figure(121)
plt.plot(time, data[:, :, 0], label='time')
plt.figure(122)
plt.plot(freq, np.abs(w_fft[:, :, 0]), label='amp')
plt.plot(freq, np.abs(amp_from_psd[:, :, 0]), label='psd_derived_amp')
plt.legend()
plt.show(block=True)



# -*- coding: utf-8 -*-
"""
@author: jundurraga-ucl
"""
from peegy.io.readers.eclipse_tools import parse_eclipse_data
from peegy.processing.pipe.io import EclipseReader
import os
# Enable below for interactive backend
# import matplotlib
# if 'qt5agg' in matplotlib.backends.backend_registry.list_builtin():
#     matplotlib.use('qt5agg')


_path = os.path.abspath(os.path.dirname(__file__))
file_path = ('/media/jundurraga/90720477720463F61/Measurements/IRU/Measurements/IRU_Soton_Manchester/'
             'Infant aided CAEP Eclipse log data/B001_S01/20160630 10-05-47_B001_S01_Block1_run1_sig1_65.txt')
data1 = parse_eclipse_data(file_name=file_path)
buffer_a = EclipseReader(file_path=file_path, buffer='A')
buffer_b = EclipseReader(file_path=file_path, buffer='B')
buffer_all = EclipseReader(file_path=file_path)
buffer_a.run()
buffer_a.plot()
buffer_b.run()
buffer_b.plot()
buffer_all.run()
buffer_all.plot()
